from .. import Rep;
from .. utils import util;

################################################################################

def read_fasta(fname, sep='auto', split=1, fieldnames=()):

    

    f = util.open_file(fname,mode='rU');
    fas = [];

    seqid = "";
    seq   = "";

    for line in f:
        line = line.strip('\n\r');
        if not line or line[0] == ">":
            if seqid:
                fas.append((seqid, seq))
            seqid = line[1:];
            seq = "";
            continue;
        seq = seq + line;

    f.close();

      # check for one last sequence
    if seq:
        fas.append((seqid,seq))

    if split:
        if sep == 'auto':
            sep_search = ['\t','|',',']
            for sep in sep_search:
                if all([sep in fas[i][0] for i in xrange(min(len(fas),100))]):
                    break
            else:
                util.warning('Could not determine FASTA separator. Please specify through sep parameter')
                sep = ''

        if sep:
            fas = [tuple(x.strip() for x in row[0].split(sep)) + row[1:] for row in fas]
            
    return Rep(fas)/fieldnames;


################################################################################





import manager
import ensure_info
from ..utils.multi_visitor import VisitorFactory, NF_ELSE
from .. import query_graph

_delay_import_(globals(),"..repops")

class CreateGraph(VisitorFactory(prefixes=("visit",), 
                                      flags=NF_ELSE), manager.Pass):
    after=set([ensure_info.EnsureInfo])
    @classmethod
    def run(cls, query, run_manager):
        self = cls()

        visited = set()
        self.queue.append(query.root)
        
        while(self.queue):
            elem = self.queue.pop()
            if(elem in visited):
                continue
            self.visit(elem)
            visited.add(elem)

        return self.graph

    def __init__(self):
        self.queue = []
        self.graph = query_graph.Graph()
   
    def visitelse(self,node):
        raise RuntimeError, "Unknown type of node encountered"
    
    def visitFixate(self,node):
        assert len(node._slices) == 1, "Query root should have one slice"
        self.graph.setRoot(node._slices[0])
        self.graph.addNode(node._slices[0])
        self.queue.append(node._slices[0])
        
    def visitExtendOp(self,node):
        if(hasattr(node,'create_graph_exec')):
            node.graph_exec(self)

    def visitDataOp(self,node):
        pass

    def visitUnaryUnaryOp(self,node):
        self.graph.addNode(node.source)
        self.graph.addEdge(query_graph.ParamEdge(node.source,node,"slice"))
        self.queue.append(node.source)
    
    def visitMultiUnaryOp(self,node):
        for pos, slice in enumerate(node.sources):
            self.graph.addNode(slice)
            self.graph.addEdge(query_graph.ParamListEdge(slice,node,"slices",pos))
        self.queue.extend(node.sources)

    def visitMultiMultiOp(self, node):
        for pos, slice in enumerate(node.sources):
            self.graph.addNode(slice)
            self.graph.addEdge(query_graph.ParamListEdge(slice,node,"slices",pos))
        self.queue.extend(node.sources)
        
    def visitFilterOp(self,node):
        self.visitUnaryUnaryOp(node)
        self.graph.addNode(node.constraint)
        self.graph.addEdge(query_graph.ParamEdge(node.constraint,node,"constraint"))
        self.queue.append(node.constraint)

   
    def visitEnsureCommonDimOp(self,node):
        self.visitUnaryUnaryOp(node)
        for slice in node.refslices:
            self.graph.addNode(slice)
            self.graph.addEdge(query_graph.ParamChoiceEdge(slice,node,"compare_slice"))
        self.queue.extend(node.refslices)
     
    def visitBroadcastOp(self,node):
        self.visitUnaryUnaryOp(node)
        for pos, slicelist in enumerate(node.refsliceslist):
            for slice in slicelist:
                self.graph.addNode(slice)
                self.graph.addEdge(query_graph.ParamChoiceListEdge(slice,node,"compare_slices",pos))
            self.queue.extend(slicelist)
       



import unittest
import numpy

from ibidas.utils import nested_array
from ibidas.utils import cutils,util
from ibidas.itypes import detector
from ibidas import *


class TestData(unittest.TestCase):
    def setUp(self):
        self.data = None

    def test_init(self):
        k = Rep(self.data)()

    def test_same(self):
        orig = Rep(self.data) 
        reload = Rep(orig()) 
        self.assertTrue(orig ==+ reload)
    
    def getRep(self):
        if(hasattr(self,'dtype')):
            return Rep(self.data,dtype=self.dtype[0])
        else:
            return Rep(self.data)
    
    def test_type(self):
        if(hasattr(self,'dtype')):
            for elem in self.dtype:
                self.assertTrue(rep(self.data,dtype=elem) ==+ Rep(self.data))

class TestScalar(TestData):
    def setUp(self):
        self.data = 1


class TestArray(TestData):
    def setUp(self):
        self.data = numpy.array([1,2,3,4,5])
        self.fres0 = 1
    
    def test_type(self):
        k = Rep(self.data)()
        self.assertTrue(k.dtype == self.data.dtype)

    def test_filter(self):
        k = Rep(self.data)
        self.assertTrue(k[k == 3]==3)
        self.assertFalse(k[k == 3] == 4)

    def test_filter2(self):
        k = Rep(self.data)
        self.assertTrue(k[0] == self.fres0)

    def test_intersect(self):
        x = Rep(self.data)
        str(x.Intersect(x))



class TestMatrix(TestArray):
    def setUp(self):
        self.data = numpy.array([[1,2,3,4,5],[1,2,3,4,5]])
        self.fres0 = numpy.array([1,1])
        self.transpose_couples = [((1,0),(1,0))]

    def test_transpose(self):
        r = self.getRep()
        pidx = range(r.Depth)
        self.assertTrue(r.Transpose(pidx) ==+ r)
       
        if(hasattr(self,'transpose_couples')):
           for i in range(len(self.transpose_couples)):
               t1,t2 = self.transpose_couples[i]
               self.assertTrue(r.Transpose(t1).Transpose(t2) ==+ r)
        
    def test_flat(self):
       f1 = Rep(self.data)
       #backward
       while(f1.Depth >= 2):
            f1 = f1.Flat(None)
       #forward
       f2 = Rep(self.data)
       while(f2.Depth >= 2):
            f2 = f2.Flat(f2._slices[0].dims[1].name)
       #all
       f3 = Rep(self.data).FlatAll()
       
       #compare
       self.assertTrue(f1 ==+ f2)
       self.assertTrue(f2 ==+ f3)
       self.assertFalse(f2 ==+ (f3.Sort()))


    def test_filter3(self):
        k = Rep(self.data)
        self.assertTrue(k[:1].Flat() ==+ k[0])

    def test_filter4(self):
        k = Rep(self.data)
        self.assertTrue(k[Rep([0],unpack=False)].Flat() ==+ k[0])

class TestNestedArray(TestMatrix):
    def setUp(self):
        self.data = util.darray([[1,2,3],[1,2],[1]])
        self.fres0 = util.darray([1,1,1])

class TestNestedMatrix(TestMatrix):
    def setUp(self):
        self.data = util.darray([[[1,2],[2,4],[3,5]],[[1,3],[5,2]],[[1,2]]])
        self.fres0 = util.darray([[1,2,3],[1,5],[1]])
        self.transpose_couples = [((0,2,1),(0,2,1)),((2,0,1),(1,2,0))]

class TestNestedMatrixString(TestMatrix):
    def setUp(self):
        self.data = util.darray([[["abc","abcd"],["def","defg"],["abg","fev"]],[["zeK","sdf"],["sdf","sdfff"]],[["sdf","kjl"]]])
        self.fres0 = util.darray([["abc","def","abg"],["zeK","sdf"],["sdf"]])
        self.transpose_couples = [((0,2,1),(0,2,1)),((2,0,1),(1,2,0))]
    
    def test_filter(self):
        k = Rep(self.data)
        self.assertTrue(k[k == "def"]=="def")
        self.assertFalse(k[k == "def"] == "defg")

class TestNestedNestedArray(TestMatrix):
    def setUp(self):
        self.data = util.darray([[[1,2],[2,3,3,3,4,5],[]],[[1],[2]],[[1,4,5]]])
        self.fres0 = util.darray([[1,2,[]],[1,2],[1]])

    def test_filter2(self):
        k = Rep(self.data)
        self.assertRaises(Exception,k[0])

    def test_filter3(self):
        pass 

    def test_filter4(self):
        pass

class TestNestedNestedMatrix(TestMatrix):
    def setUp(self):
        self.data = util.darray([[[[1,2],[3,4]],[[5,6],[7,9],[9,8]],[[4,9]]],[[[1,4]],[[2,5]]],[[[1,5],[4,7],[8,5]]]])
        self.transpose_couples = [((0,1,3,2),(0,1,3,2)),((0,3,1,2),(0,2,3,1)),((3,0,1,2),(1,2,3,0))]

    def test_filter2(self):
        k = Rep(self.data)
        self.assertRaises(Exception,k[3])

class TestArrayNestedNestedMatrix(TestMatrix):
    def setUp(self):
        nnmatrix = util.darray([[[[1,2],[3,4]],[[5,6],[7,9],[9,8]],[[8,9]]],[[[1,4]],[[2,5]]],[[[1,5],[4,7],[8,5]]]])
        self.data = util.darray([nnmatrix, nnmatrix, nnmatrix])
        self.transpose_couples = [((1,2,3,4,0),(4,0,1,2,3)),((1,0,2,3,4),(1,0,2,3,4)),((1,2,0,3,4),(2,0,1,3,4)),((1,2,3,0,4),(3,0,1,2,4))]

    def test_filter2(self):
        k = Rep(self.data)
        self.assertRaises(Exception,k[3])
    
    def test_flat(self):
       f1 = Rep(self.data)
       #backward
       while(f1.Depth >= 3):
            f1 = f1.Flat(-1)
       #forward
       f2 = Rep(self.data)
       while(f2.Depth >= 3):
            f2 = f2.Flat(f2._slices[0].dims[2].name)
       #all
       
       #compare
       self.assertTrue(f1 ==+ f2)

class TestNestedVarMatrix(TestMatrix):
    def setUp(self):
        self.data = util.darray([[[1,2],[3,4],[5,6]],[[4,5,6],[7,8,9]]])
        self.dtype=["[arr]<[var1:.]<[var2:*.]<int64"]
        self.fres0 = util.darray([[1,3,5],[4,7]])
        self.transpose_couples = [((0,2,1),(0,2,1))]
        
if __name__ == "__main__":
    unittest.main()



import wrapper_tsv
_delay_import_(globals(),"wrapper_py","rep")

def read_phospho(path):
    fpath = path + 'kinase_all.txt' 
    kinase_list = wrapper_tsv.read_tsv(fpath)
    kinase_list = kinase_list[kinase_list.map(lambda x:x[0] != "#", otype="bool")]

    data = []
    dtype = "[kinase_binding:*](tuple(kinase=string[], orf=string[], score=real64))"
    for elem in kinase_list():
        filename = path + str(elem) + ".txt"
        file = open(filename)
        file.next()
        for row in file:
            r = row.split('\t')
            data.append((str(elem), r[0], float(r[1])))

    return Rep(data,dtype)



import __builtin__

class Context(object):
    """Context objects store all actions on them, which
    can then be replayed on another object with the _apply function below"""

    __slots__ = ['_action','_prev']
    __aterror__ = set(('__array_struct__','__array__','__array_interface__',\
                        '__array_priority__','__array_wrap__','trait_names','_getAttributeNames'))

    def __init__(self,prev,*action):
            self._prev = prev
            self._action = action

    def __getattr__(self,name):
        #some support for numpy and ipython
        if(name in self.__aterror__):
            raise AttributeError
        r = Context(self,'DOT',name)
        return r

        #return Context(self,'DOT',name)

    def __getitem__(self,key):
        return Context(self,'ITEM',key)

    def __repr__(self):
        return repr(self._prev) + str(self._action)

    def __call__(self,*params,**kwds):
        return Context(self,'CALLSELF',params,kwds)
   
    def __iter__(self):
        return Context(self,"ITER")

    def next(self):
        return Context(self,"NEXT")

    def __neg__(self):
        return Context(self,'UNOP','__neg__')

    def __pos__(self):
        return Context(self,'UNOP','__pos__')
    
    def __invert__(self):
        return Context(self,'UNOP','__invert__')

    def __add__(self, other):
        return Context(self,'BINOP','__add__',other)

    def __radd__(self, other):
        return Context(self,'BINOP','__radd__',other)
    
    def __sub__(self, other):
        return Context(self,'BINOP','__sub__',other)

    def __rsub__(self, other):
        return Context(self,'BINOP','__rsub__',other)

    # multiplication ( * )
    def __mul__(self, other):
        return Context(self,'BINOP','__mul__',other)
    
    def __rmul__(self, other):
        return Context(self,'BINOP','__rmul__',other)

    # division ( / )
    def __div__(self, other):
        return Context(self,'BINOP','__div__',other)
    
    def __rdiv__(self, other):
        return Context(self,'BINOP','__rdiv__',other)
    
    # floor division ( // )
    def __floordiv__(self, other):
        return Context(self,'BINOP','__floordiv__',other)
    
    def __rfloordiv__(self, other):
        return Context(self,'BINOP','__rfloordiv__',other)
    
    # power (**)
    def __pow__(self, other):
        return Context(self,'BINOP','__pow__',other)
    
    def __rpow__(self, other):
        return Context(self,'BINOP','__rpow__',other)
    
    # modulo (%)
    def __mod__(self, other):
        return Context(self,'BINOP','__mod__',other)
    
    def __rmod__(self, other):
        return Context(self,'BINOP','__rmod__',other)
 
    # and operator ( & )
    def __and__(self, other):
        return Context(self,'BINOP','__and__',other)
    
    def __rand__(self, other):
        return Context(self,'BINOP','__rand__',other)
    
    # or operator ( | )
    def __or__(self, other):
        return Context(self,'BINOP','__or__',other)

    def __ror__(self, other):
        return Context(self,'BINOP','__ror__',other)
     
    # exclusive-or operator ( ^ )
    def __xor__(self, other):
        return Context(self,'BINOP','__xor__',other)

    def __rxor__(self, other):
        return Context(self,'BINOP','__rxor__',other)

    # less-than ( < )
    def __lt__(self, other):
        return Context(self,'BINOP','__lt__',other)

    # less-than-or-equals ( <= )
    def __le__(self, other):
        return Context(self,'BINOP','__le__',other)

    # equals ( == )
    def __eq__(self, other):
        return Context(self,'BINOP','__eq__',other)

    # not-equals ( != )
    def __ne__(self, other):
        return Context(self,'BINOP','__ne__',other)

    # greater-than ( > )
    def __gt__(self, other):
        return Context(self,'BINOP','__gt__',other)

    # greater-than-or-equals ( >= )
    def __ge__(self, other):
        return Context(self,'BINOP','__ge__',other)

    def _call(self, func, *params, **kwds):
        return Context(self, "APPLYFUNC", func, params, kwds)

def _apply(ct, obj, **extra_params):
    """Apply a context ct to an object obj"""
    result = obj
    actionlist = []
    while(not ct._prev is None):
        actionlist.append(ct._action)
        ct = ct._prev
   
    for action in actionlist[::-1]:
        if(action[0] == 'DOT'):
            result = getattr(result,action[1])
        elif(action[0] == 'ITEM'):
            result = result[action[1]]
        elif(action[0] == 'CALLSELF'):
            param = []
            for p in action[1]:
                #if(isinstance(p,Context)):
                #    param.append(_apply(p, obj, **extra_params))
                #else:
                param.append(p)
            kwds = dict()
            for k,v in action[2].iteritems():
                #if(isinstance(v,Context)):
                #    kwds[k] = _apply(v, obj, **extra_params)
                #else:
                kwds[k] = v

            if(extra_params):
                if(result.func_name in extra_params):
                    kwds.update(extra_params[result.func_name])
            result = result(*param, **kwds)
        elif(action[0] == 'BINOP'):
            if(isinstance(action[2],Context)):
                other = _apply(action[2], obj, **extra_params)
            else:
                other = action[2]
            result = getattr(result,action[1])(other)
        elif(action[0] == 'UNOP'):
            result = getattr(result,action[1])()
        elif(action[0] == 'SETITEM'): #cannot be recorded, but can be played
            if(isinstance(action[2],Context)):
                value = _apply(action[2], obj, **extra_params)
            else:
                value = action[2]
            result.__setitem__(action[1],action[2])
        elif(action[0] == 'APPLYFUNC'):
            param = []
            for p in action[2]:
                if(isinstance(p,Context)):
                    param.append(_apply(p, obj, **extra_params))
                else:
                    param.append(p)
            kwds = action[3].copy()
            for k,v in action[3].iteritems():
                if(isinstance(v,Context)):
                    kwds[k] = _apply(v, obj, **extra_params)
            if(extra_params):
                if(action[1].func_name in extra_params):
                    kwds.update(extra_params[action[1].func_name])
            result = action[1](result,*param, **kwds)
    return result


_=Context(None)
__builtin__.__dict__['_'] = _



import manager
import create_graph
import wrapper_planner
import serialize_exec

from ..utils import util
from ..utils.multi_visitor import VisitorFactory, NF_ERROR, F_CACHE
from ..constants import *
_delay_import_(globals(),"..query_graph")
_delay_import_(globals(),"..ops")
_delay_import_(globals(),"..itypes","dimpaths")

class PythonPeepHole(VisitorFactory(prefixes=("visit",), flags=NF_ERROR), manager.Pass):
    after=set([wrapper_planner.WrapperPlanner])
    before=set([serialize_exec.SerializeExec])

    @classmethod
    def run(cls, query, run_manager):
        self = cls()
        self.graph = run_manager.pass_results[create_graph.CreateGraph]

        for node in list(self.graph.nodes):
            self.visit(node)
    
    def visitNode(self, node):
        pass

    def visitBroadcastOp(self, node):
        if not node.partial:
            return
        nplans = dimpaths.processPartial(node.bcdims, node.plan)
        updated=False
        bcpos = 0
        for oplan,nplan in zip(node.plan, nplans):
            if oplan != nplan:
                self.remove_compare_slice(node,bcpos)
                updated=True
            elif oplan ==  BCEXIST:
                bcpos += 1
        if updated:
            if bcpos == 0:
                self.remove_unaryop(node) 
            else:
                node = self.graph.copyNode(node)
                node.plan = nplans

    def remove_compare_slice(self, node, pos):
        for edge in list(self.graph.edge_target[node]):
            if edge.name == "compare_slices":
                if edge.pos == pos:
                    self.graph.dropEdge(edge)
                elif edge.pos >= pos:
                    edge.pos -= 1

            
    def remove_unaryop(self, node):
        source = self.graph.getDataEdge(node).source
        target_edges = list(self.graph.edge_source[node])
        self.graph.dropNode(node)
        for target_edge in target_edges:
            assert isinstance(target_edge,query_graph.ParamEdge), "Unknown edge type encountered"
            target_edge.source = source
            self.graph.addEdge(target_edge)





import SimpleXMLRPCServer
from utils import util
from representor import Representor
import thread

def Serve(portnumber=9889):
    """Starts xml-rpc server on the supplied port number. Can be used to contact ibidas
       from other applications.
       
       XML-RPC interface:

       * startSession() (optional): create session namespace, returns sessionid

       * closeSession(sessionid): closes specified session

       * query(query, sessionid=-1): executes query string in session, returns result.
       
       * execute(query, sessionid=-1): exexutes query string in session

       * set(name, val, sessionid=-1): sets variable in session
        
       The default session id of -1 gives direct contact with the ibidas command line session

       Starting the server::
           
           $ ibidas
           >>> s = Serve()
           >>> z = Rep([1,2,3,4])
      
       Connecting to ibidas (python, but can be any other application supporting xml-rpc)::
           $ipython
           >>> import xmlrpclib
           >>> s = xmlrpclib.ServerProxy("http://localhost:9889")
           >>> s.query("z + 4")
           [4,5,6,7]

       """

    s = SimpleXMLRPCServer.SimpleXMLRPCServer(("localhost",portnumber),allow_none=True,bind_and_activate=False, logRequests=False)
    s.allow_reuse_adress = True
    s.server_bind()
    s.server_activate()
    l = LocalRequestHandler()
    s.register_instance(l,allow_dotted_names=False)
    s.register_introspection_functions()
    s.register_multicall_functions()
    thread.start_new_thread(runServer,(s,))
    return s

def runServer(s):
    s.serve_forever()


class LocalRequestHandler(object):
    def __init__(self):
        self.sessions = []

    def startSession(self):
        if None in self.sessions:
            sessionid = self.sessions.index(None)
        else:
            sessionid = len(self.sessions)
            self.sessions.append(None)
        self.sessions[sessionid] = {}
        return sessionid

    def closeSession(self, sessionid):
        assert isinstance(sessionid,int), "Session id should be an int"
        assert sessionid >= 0 and sesionid < len(self.sessions), "Session id not valid"
        self.sessions[sessionid] = None

    def query(self, query, sessionid=-1):
        assert isinstance(sessionid,int), "Session id should be an int"
        assert sessionid >= -1 and sessionid < len(self.sessions), "Session id not valid"
      
        session = getSession(self, sessionid)

        res = eval(query,getGlobals(), session)
        if(isinstance(res,Representor)):
            res = res.ToPython()
        
        return res

    def execute(self, statement, sessionid=-1):
        assert isinstance(sessionid,int), "Session id should be an int"
        assert sessionid >= -1 and sessionid < len(self.sessions), "Session id not valid"

        session = getSession(self, sessionid)
       
        exec statement in getGlobals(), session

        return True

    def set(self, name, value, sessionid=-1):
        session = getSession(self, sessionid)
        session[name] = value


def getGlobals():
    if '__IPYTHON__active' in globals()['__builtins__']:
        return globals()['__builtins__']['__IPYTHON__'].user_global_ns
    else:
        return globals()

def getSession(rh, sessionid):
    if(sessionid == -1):
        assert '__IPYTHON__active' in globals()['__builtins__'], "Ipython not active, cannot find shell session"
        session = globals()['__builtins__']['__IPYTHON__'].user_ns
    else:
        session = rh.sessions[sessionid]
    return session




"""Classes to handle array data with sparse values"""


import numpy
import operator
import array
from itertools import izip_longest

from missing import Missing
import util
_delay_import_(globals(),"cutils")
#_delay_import_(globals(),"ibidas.itypes","Missing")



class FullSparse(numpy.ndarray):
    def __new__(cls,obj):
        res = numpy.array(obj).view(cls)
        return res

    def getVal(self):
        return self.view(numpy.ndarray)
    val = property(fget=getVal)
    
    def _getClasses(self):
        classes = set()
        if(self.dtype == object):
            classes |= set([value.__class__ for value in self.ravel()])
        else:
            classes.add(self.dtype.type)

        return classes
    classes=property(fget=_getClasses)

    def map(self, func, **kwargs):
        out_empty = kwargs.get('out_empty',Missing)
        
        if(kwargs.get('has_missing',True)):
            func = elem_empty_filter(func,out_empty)
        
        otype = kwargs.get('otype',self.dtype)

        if(otype == object):
            res = util.darray([func(elem) for elem in self.ravel()])
            res.shape = self.shape
        elif(self.dtype.char == 'S' or self.dtype.char == 'U' or self.dtype.char == 'V'):
            res = util.darray([func(elem) for elem in self.ravel()],otype)
            res.shape = self.shape
        else:
            func = numpy.frompyfunc(func,1,1)
            res = numpy.cast[otype](func(self))
        
        return FullSparse(res)

    def filter_tomissing(self,filter):
        nval = self.copy()
        nval = numpy.cast[object](nval)
        nval[~filter] = Missing
        return nval.view(FullSparse)

    def hasSparse(self):
        return numpy.equal(self,Missing).any()

    def replace_missing(self,empty_replace=Missing,otype=None):
        if(not empty_replace is Missing):
            missing_filter = numpy.equal(self,Missing)
            res = self.copy()
            res[missing_filter] = empty_replace
        else:
            res = self
        if(not otype is None):
            res = numpy.cast[otype](res)
        return res
    
    def without_missing(self):
        val = self.copy()
        missing_filter = numpy.equal(self, Missing)
        return val[~missing_filter]


    def min(self,has_missing=True,out_empty=Missing):
        if(has_missing):
            val = self.without_missing()
            if len(val) > 0:
                return val.val.min()
            else:
                return out_empty
        else:
            return self.val.min()

    def max(self,has_missing=True,out_empty=Missing):
        if(has_missing):
            val = self.without_missing()
            if len(val) > 0:
                return val.val.max()
            else:
                return out_empty
        else:
            return self.val.max()
    
    def all(self,has_missing=True,out_empty=Missing):
        if(has_missing):
            val = self.without_missing()
            if len(val) > 0:
                return val.val.all()
            else:
                return out_empty
        else:
            return self.val.all()
    
    def any(self,has_missing=True,out_empty=Missing):
        if(has_missing):
            val = self.without_missing()
            if len(val) > 0:
                return val.val.any()
            else:
                return out_empty
        else:
            return self.val.any()




def sparse_concat(*seqs):
    poss = []
    vals = []
    length = 0
    if not seqs:
        return PosSparse([],[],0)
    

    if all([isinstance(seq,numpy.ndarray) for seq in seqs]):
        for seq in seqs:
            if(isinstance(seq,FullSparse)):
                pass
        return numpy.concatenate(seqs) 
    else:
       pass 

    for seq in seqs:
        if not isinstance(seq, PosSparse):
            seq = from_seq(seq)
        assert isinstance(seq,PosSparse), "Cannot convert input to sparse list"
        poss.append(seq.sorted_pos + length)
        length += seq.length
        vals.append(seq.val)

    return PosSparse(numpy.concatenate(vals), numpy.concatenate(poss), shape)

def elem_empty_filter(func,out_empty=Missing):
    def f(elem):
        if elem is Missing:
            return out_empty
        else:
            return func(elem)
    return f

def _inbetween_tile(lpos,apos,nrows,ncols,times):
    steps = (numpy.arange(nrows)+1) * ncols
    templpos = []
    tempapos = []
    steppos = 0
    pospos = 0
    lenlpos = len(lpos)
    while(steppos < nrows and pospos < lenlpos):
        curstepend = steps[steppos]
        beginpospos = pospos
        while(pospos < lenlpos and apos[pospos] < curstepend):
            pospos+=1
        t = apos[beginpospos:pospos]
        tempapos.extend(numpy.tile(t,times) + numpy.repeat(numpy.arange(times) * ncols,len(t)) + steppos * ncols * (times-1))
        t = lpos[beginpospos:pospos]
        templpos.extend(numpy.tile(t,times))
        steppos += 1

    lpos = numpy.cast[int](templpos)
    apos = numpy.cast[int](tempapos)
    return (lpos,apos)





from xml.dom.minidom import parse
import os

import wrapper
from ..itypes import rtypes
from .. import repops_slice, ops
from ..constants import *

import python
from ..itypes import detector
from ..utils import nested_array, util

class XMLRepresentor(wrapper.SourceRepresentor):
    def __init__(self, filename):
        dom = parse(filename)
        allowed_nodes = set([dom.ELEMENT_NODE])
        res = parse_node(dom, allowed_nodes)
        if isinstance(res,list) and len(res) == 1:
            res = res[0]

        det = detector.Detector()
        det.process(res)
        dtype = det.getType()

        slice = XMLOp(res, dtype, "data")
        if(slice.type.__class__ is rtypes.TypeRecordDict):
            nslices = repops_slice.UnpackTuple._apply(slice)
        else:
            nslices = [slice]
        self._initialize(tuple(nslices))


class XMLOp(ops.ExtendOp):
    __slots__ = ["data"]

    def __init__(self, data, rtype, name):
        self.data = data
        ops.ExtendOp.__init__(self,name=name,rtype=rtype)

    def py_exec(self):
        ndata = nested_array.NestedArray(self.data,self.type)
        return python.ResultOp.from_slice(ndata,self)


def parse_node(node, allowed_nodes):
    if node.nodeType in allowed_nodes:
        res = {'node_name':node.nodeName}
        if node.hasChildNodes():
            res['node_children'] = [parse_node(n,allowed_nodes) for n in node.childNodes if n.nodeType in allowed_nodes]
        if node.hasAttributes():
            res.update(dict(node.attributes.items()))
        if not node.nodeValue is None and not (isinstance(node.nodeValue,str) and node.nodeValue.strip() == ""):
            res['node_value'] = node.nodeValue
        return res 
    else:
        return [parse_node(n, allowed_nodes) for n in node.childNodes if n.nodeType in allowed_nodes]



import sys
import getopt
import os

#make sure we import the development version
try:
    sys.path = [os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)),'..'))] + sys.path
except:
    pass

from ibidas import *
from ibidas.server import Serve
from ibidas import representor

import IPython

ipversion = IPython.__version__.split('.')
oldip = int(ipversion[0]) == 0 and int(ipversion[1]) <= 10

if oldip:
    from IPython.Shell import IPShellEmbed
    from IPython.ipapi import TryNext
    from IPython.genutils import dir2
    import IPython.rlineimpl as readline
else:   
    from IPython.config.loader import Config
    from IPython.frontend.terminal.embed import InteractiveShellEmbed
    
    #from IPython.ipapi import TryNext
    #from IPython.genutils import dir2
    #import IPython.rlineimpl as readline

from ibidas.command_parser import parseCommandLine

import numpy
if numpy.version.version == '1.6.0':
    print 'WARNING: numpy 1.6.0 has a regression (http://projects.scipy.org/numpy/ticket/1870). Some commands wil not work. Please update to a newer version.'
    print ''

nv = numpy.version.version.split('.')
if len(nv) ==3:
    try:
        if int(nv[0]) == 1 and (int(nv[1]) <= 3 or (int(nv[1]) == 4 and int(nv[2]) == 0)):
            print 'WARNING: your numpy version (%s) is too old. Some commands will not work. Please update to a newer version.' % numpy.version.version
            print ''
    except ValueError:
        pass

rc_path = "~/.ibidas_rc"
rc_path = os.path.expanduser(rc_path)
global ipshell
global _rep_res
_rep_res = None

def complete_show(line, matches, longest_match):
    global _rep_res
    ipshell.IP.write('\n')
    if _rep_res is None:
        ipshell.IP.write(str(matches))
        ipshell.IP.write('\n')
    else:
        contextobj, attr, words = _rep_res
        _rep_res = None
        slice_names = contextobj.Names
        slice_names_set = set(slice_names)
        
        rem_words = []
        pos_slices = []
        for word in words:
            if word in slice_names_set:
                pos_slices.append(slice_names.index(word))
            else:
                rem_words.append(word)
        if(pos_slices):
            pos_slices.sort()
            ipshell.IP.write(contextobj.Get(*pos_slices).Info.info)
            ipshell.IP.write('\n')
        if(rem_words):
            ipshell.IP.write(str(rem_words))
            ipshell.IP.write('\n')
        _rep_res = None
        
   
    #well, this part was well documented, not....
    realline = readline.get_line_buffer()
    ipshell.IP.interact_prompt()
    ipshell.IP.write(realline[:readline.get_endidx()])
    addline = realline[readline.get_begidx() + len(line):]
    addline += chr(8) * len(addline)
    ipshell.IP.write(addline)



def rep_completer(self,event, line = None):
    global _rep_res
    if line is None:
        line = readline.get_line_buffer()
    try:
        contextobj, attr = parseCommandLine(line, ipshell)
        if(not isinstance(contextobj, representor.Representor)):
            raise TryNext
        words = dir2(contextobj)
        if attr:
            words = [word for word in words if word.startswith(attr)]
        else:
            words = [word for word in words if not word.startswith("_")]
        _rep_res = (contextobj, attr, words)
        if attr:
            words = [event.symbol[:-len(attr)] + word for word in words]
        else:
            words = [event.symbol + word for word in words]
        return words
    except: 
        raise TryNext


if oldip:
    ipshell = IPShellEmbed(argv=sys.argv[1:],banner='Welcome to the IBIDAS system',exit_msg='IBIDAS shutting down ...',rc_override={'cache_size':0, 'readline_omit__names':2})
    del ipshell.IP.user_ns['_']
    ipshell()
else:
    #Hack to get 'In' to show in Ipython, instead of the builtin 'In' history.
    #Requires a) setting builtin In, b) user_ns to globals(), c) del ipshell.user_ns['In']
    #Probably a bit fragile..
    import __builtin__
    __builtin__.__dict__['In'] = In
    cfg = Config()
    cfg.InteractiveShellEmbed.cache_size = 0
    #rc_override={'cache_size':0, 'readline_omit__names':2}
    ipshell = InteractiveShellEmbed(config=cfg, user_ns=globals(), banner2='Welcome to the IBIDAS system', exit_msg='IBIDAS shutting down ...')
    del ipshell.user_ns['_']
    del ipshell.user_ns['In']
    ipshell()

#ipshell.IP.set_hook('complete_command', rep_completer, re_key = '.*')

#if(ipshell.IP.has_readline):
#    readline.set_completion_display_matches_hook(complete_show)



import operator
from itertools import chain
from constants import *
import repops
import ops


_delay_import_(globals(),"representor")
_delay_import_(globals(),"wrappers","python")
_delay_import_(globals(),"utils","util","context")
_delay_import_(globals(),"itypes","rtypes","dimpaths","casts")

class ProjectDim(repops.UnaryOpRep):
    def _sprocess(self, source, name):
        nslices = [slice for slice in source._slices if slice.dims.hasName(name)]
        if not nslices:
            raise AttributeError, "Cannot find matching slices for dimension '" + str(name) + "'"
        self._initialize(tuple(nslices))

class ProjectBookmark(repops.UnaryOpRep):
    def _sprocess(self, source, name):
        nslices = [slice for slice in source._slices if name in slice.bookmarks]
        if not nslices:
            raise AttributeError, "Cannot find matching slices for bookmark '" + str(name) + "'"
        self._initialize(tuple(nslices))

class RequestUnaryOpRep(repops.UnaryOpRep):
    def __init__(self,source,*args, **kwds):
        self._req_sources = []
        repops.UnaryOpRep.__init__(self,source,*args, **kwds)


class To(RequestUnaryOpRep):
    def _sprocess(self, source, *slicesel, **kwargs):
        do = kwargs.pop("Do")
        assert not kwargs, "Unknown parameters: " + str(kwargs)

        if not self._req_sources:
            for ssel in slicesel:
                r = source.Get(ssel) 
                #assert len(r._slices) == 1, "To action can only be applied to single slices"
                all_pos = []
                for slice in r._slices:
                    #slice = r._slices[0]
                    assert slice in source._slices, "Selected slice in to should not be operated upon"
                    pos = source._slices.index(slice)
                    all_pos.append(pos)

                if(isinstance(do, context.Context)):
                    r = context._apply(do, r)
                else:
                    r = do(r)

                self._req_sources.append((all_pos, r))

        nslices = list(source._slices)
        for all_pos, r in self._req_sources:
            if not r._slicesKnown():
                return 
            for slice,pos in zip(r._slices, all_pos):
                nslices[pos] = slice

        self._initialize(tuple(nslices))
       

class Project(RequestUnaryOpRep):
    def __init__(self,source,*args, **kwds):
        self._req_sources = []
        repops.UnaryOpRep.__init__(self,source,*args, **kwds)

    def _getUsedSourceSlicesSet(self,req_sources):
        nslices = []
        for name, elem in req_sources:
            if isinstance(elem, representor.Representor):
                nslices.extend(elem._slices)
            else:
                nslices.extend(elem)
        return nslices

    def _sprocess(self, source, *args, **kwds):
        cur_slices = self._source._slices
        
        nslices = []
        if(not self._req_sources):
            req_sources = []
            for name, elem in chain(zip([None] * len(args),args),kwds.iteritems()):
                if(isinstance(elem, context.Context)):
                    elem = context._apply(elem, self._source)
                elif(isinstance(elem, basestring)):
                    if(elem == "~"):
                        used_slices = set([slice.name for slice in self._getUsedSourceSlicesSet(req_sources)])
                        nelem = [slice for slice in cur_slices if slice.name not in used_slices]
                    elif(elem == "#"):
                        common_dims = set([slice.dims for slice in cur_slices])
                        if len(common_dims) != 1:
                            raise RuntimeError, "Cannot use # selector as fields do not have a common dimension"
                        nelem = cur_slices[:1]
                    elif(elem == "*"):
                        nelem = cur_slices
                    else:
                        nelem = [slice for slice in cur_slices if slice.name == elem]
                        if(not nelem):
                            if len(cur_slices) == 1 and isinstance(cur_slices[0].type,rtypes.TypeTuple) \
                               and (elem in cur_slices[0].type.fieldnames or (isinstance(elem,int) and elem >= 0 and elem < len(cur_slices[0].type.fieldnames))):
                               nelem = UnpackTuple._apply(cur_slices[0],elem)
                            else:
                               nelem = [slice for slice in cur_slices if elem in slice.bookmarks]
                        
                        if not nelem:
                            raise AttributeError, "Cannot find attribute '" + elem + "'"
                    elem = nelem
                elif(isinstance(elem, representor.Representor)):
                    pass
                elif(isinstance(elem, tuple)):
                    elem = Tuple(self._source.Get(*elem))
                elif(isinstance(elem, list)):
                    if(len(elem) == 1):
                        elem = self._source.Get(*elem).Array()
                    else:
                        elem = self._source.Get(*elem).Array()
                else:
                    elem = util.select(cur_slices, elem)
                   
                req_sources.append((name,elem))
            self._req_sources = req_sources
     
        
        for name,elem in self._req_sources:
            if(isinstance(elem, representor.Representor)):
                if not elem._slicesKnown():
                    return 
                elem = [slice for slice in elem._slices]

            if(name):
                if not len(elem) == 1:
                    raise AttributeError, "Cannot find unique slice for '" + name + "', found: " + str(elem)
                nslices.append(ops.ChangeNameOp(elem[0],name))
            else:
                nslices.extend(elem)
        
        if not nslices:                
            raise AttributeError,  "No slices found with: " + str(args) + " and " + str(kwds)
        
        return self._initialize(tuple(nslices)) 
    

class Unproject(Project):
    def _sprocess(self, source, *args, **kwds):
        selslices = Project(source, *args, **kwds)._slices
        nslices = [slice for slice in source._slices if not slice in selslices]
        return self._initialize(tuple(nslices)) 


class AddSlice(repops.UnaryOpRep):
    def _sprocess(self, source, data, name=None, dtype=None):
        if not dtype is None:
            refdims = sum([slice.dims for slice in source._slices],tuple())
            dtype = rtypes.createType(dtype,refdims=refdims)
            
        data = python.Rep(data, dtype=dtype, name=name)

        nslices = list(source._slices)
        nslices.extend(data._slices)
        return self._initialize(tuple(nslices)) 
       
class UnpackTuple(repops.UnaryOpRep):
    def _process(self,source,name=None,unpack=True):
        """
        Parameters:
        source: source to unpack active slices from
        name: (Optional) if given, unpack only fields with this name
                if not given, unpack all fields from tuple.
        """
        if not source._typesKnown():
            return

        if not len(source._slices) == 1:                
            if(name):
                raise AttributeError, "Asked to unpack tuple attribute '" + \
                    name + "', but cannot find a tuple."
            else:
                raise AttributeError, "No tuple to unpack"
        slice = source._slices[0]
        nslices = self._apply(slice,name,unpack=unpack)
        return self._initialize(tuple(nslices))

    @classmethod
    def _apply(cls, slice, name=None, unpack=True):
        if(not isinstance(slice.type, rtypes.TypeTuple)):
            if(name):
                raise AttributeError, "Asked to unpack tuple attribute '" + \
                    name + "', but cannot find a tuple"
            else:
                raise AttributeError, "No tuple to unpack"

        if(name is None):
            nslices = [ops.UnpackTupleOp(slice, idx) for idx in range(len(slice.type.subtypes))]
        else: 
            try:
                idx = int(name)
            except ValueError:
                assert isinstance(name, basestring), \
                            "Tuple slice name should be a string"
                try:
                    idx = slice.type.fieldnames.index(name)
                except ValueError:
                    raise AttributeError, "Cannot unpack, as tuple has no field '" + str(name) + "'"
            nslices = [ops.UnpackTupleOp(slice, idx)]
        
        if(unpack):
            for pos, nslice in enumerate(nslices):
                while(nslice.type.__class__ is rtypes.TypeArray):
                    nslice = ops.UnpackArrayOp(nslice)
                nslices[pos] = nslice
        return nslices

class Bookmark(repops.UnaryOpRep):
    def _sprocess(self, source, *names, **kwds): #{{{
        nslices = source._slices
        if(len(names) == 1):
            nslices = [ops.ChangeBookmarkOp(slice,names[0]) for slice in nslices]
        elif(len(names) > 1):
            unique_first_dims = util.unique([slice.dims[0] for slice in source._slices])
            assert (len(names) == len(unique_first_dims)), \
                "Number of new slice names does not match number of slices"
            for name,dim in zip(names,unique_first_dims):
                kwds[dim.name] = dim

        if(kwds):
            nnslices = []
            for slice in source._slices:
                if(slice.dims[0].name in kwds):
                    nslice = slice.ChangeBookmarkOp(slice,kwds[slice.dims[0].name])
                else:
                    nslice = slice
                nnslices.append(nslice)
            nslices = nnslices
                
        return self._initialize(tuple(nslices))
        #}}}


class SliceRename(repops.UnaryOpRep):
    def _sprocess(self, source, *names, **kwds): #{{{
        if(names):
            assert (len(names) == len(source._slices)), \
                "Number of new slice names does not match number of slices"
            nslices = [ops.ChangeNameOp(slice,name) 
                    for slice, name in zip(source._slices, names)]
        else:
            nslices = list(source._slices)
            for name, newname in kwds.iteritems():
                sels = [(pos,slice) for pos, slice in enumerate(source._slices) if slice.name == name]
                if isinstance(newname,tuple):
                    assert len(newname) == len(sels), 'Number of new names for slice %s does not match number of slices (%d)' % (name, len(sels))
                    for (pos,slice),nname in zip(sels,newname):
                        nslices[pos] = ops.ChangeNameOp(slice, nname)
                else:
                    for (pos,slice) in sels:
                        nslices[pos] = ops.ChangeNameOp(slice, newname)

        return self._initialize(tuple(nslices))
        #}}}

class SliceCast(repops.UnaryOpRep):
    def _process(self, source, *newtypes, **kwds): #{{{
        if not source._typesKnown():
            return
            
        if(newtypes):
            if len(newtypes) > 1:
                assert (len(newtypes) == len(source._slices)), \
                    "Number of new slice types does not match number of slices"
                nslices = [ops.CastOp(slice,rtypes.createType(newtype)) 
                        for slice, newtype in zip(source._slices, newtypes)]
            else:
                nslices = [ops.CastOp(slice,rtypes.createType(newtypes[0])) 
                        for slice in source._slices]
        else:
            nslices = list(source._slices)
            for name, newtype in kwds.iteritems():
                sels = [(pos,slice) for pos, slice in enumerate(source._slices) if slice.name == name]
                if isinstance(newtype,tuple):
                    assert len(newtype) == len(sels), 'Number of new types for slice %s does not match number of slices (%d)' % (name, len(sels))
                    for (pos,slice),ntype in zip(sels,newtype):
                        ntype = rtypes.createType(ntype)
                        nslices[pos] = ops.CastOp(slice, ntype)
                else:
                    for (pos,slice) in sels:
                        rnewtype = rtypes.createType(newtype)
                        nslices[pos] = ops.CastOp(slice, rnewtype)

        return self._initialize(tuple(nslices))
        #}}}


class Tuple(repops.UnaryOpRep):
    _ocls = ops.PackTupleOp
    def _sprocess(self, source, **kwargs):
        nslice = self._apply(source._slices, **kwargs)
        #initialize object attributes
        return self._initialize((nslice,))

    @classmethod
    def _apply(cls, slices, to_python=False, **kwargs):
        cdimpath = dimpaths.commonDimPath([slice.dims for slice in slices])
        nslices = []
        for slice in slices:
            oslice = slice
            if(to_python):
                while(len(slice.dims) > len(cdimpath)):
                    slice = ops.PackListOp(slice)
            else:
                if(len(slice.dims) > len(cdimpath)):
                    slice = ops.PackArrayOp(slice, ndim=len(slice.dims) - len(cdimpath))
            nslices.append(slice)
    
        return cls._ocls(nslices, **kwargs)
       
class Dict(Tuple):
    _ocls = ops.PackDictOp

class IndexDict(Tuple):
    _ocls = ops.PackIndexDictOp

class HArray(repops.UnaryOpRep):
    def _process(self, source, name=None):
        if not source._typesKnown():
            return
        
        #commonify dimensions
        cdimpath = dimpaths.commonDimPath([slice.dims for slice in source._slices])
        nslices = []
        for slice in source._slices:
            oslice = slice
            if(len(slice.dims) > len(cdimpath)):
                slice = ops.PackArrayOp(slice, ndim=len(slice.dims) - len(cdimpath))
            nslices.append(slice)

        #cast to common type
        ntype = casts.castMultipleImplicitCommonType(*[slice.type for slice in nslices])
        nnslices = []
        for slice in nslices:
            if(ntype != slice.type):
                slice = ops.CastOp(slice,ntype)
            nnslices.append(slice)
    
        nslice = ops.HArrayOp(nnslices,name=name)
        nslice = ops.UnpackArrayOp(nslice)

        #initialize object attributes
        return self._initialize((nslice,))



class ToPythonRep(repops.UnaryOpRep):
    def _process(self, source):
        if not source._typesKnown():
            return

        nslices = []
        for slice in source._slices:
            if slice.type._requiresRPCconversion():
                slice = ops.ToPythonOp(slice)
            nslices.append(slice)
       
        if(len(nslices) > 1):
            nslice = Tuple._apply(nslices,to_python=True)
        else:
            nslice = nslices[0]

        while(nslice.dims):
            nslice = ops.PackListOp(nslice)

        return self._initialize((nslice,))



        



import copy
import wrapper
import operator
import sqlalchemy
from sqlalchemy.sql import expression as sql
from collections import defaultdict

from .. import ops
from ..constants import *
from ..utils.multi_visitor import VisitorFactory, NF_ELSE, F_CACHE, NF_ERROR
from ..utils import util

from ..passes import manager, create_graph, annotate_replinks, serialize_exec
from .. import query_graph


from ..utils import nested_array
from ..utils.missing import Missing
from ..itypes import rtypes, dimensions, dimpaths
import python
from .. import repops_slice

alias_name_id = util.seqgen()

class TypeMapperToSQLAlchemy(VisitorFactory(prefixes=("convert",), #{{{
                                      flags=NF_ELSE)):
    
    def convertTypeAny(self, rtype):
        return False
    
    def convertTypeUInt64(self, rtype): #to be checked: can databases handle uint64?
        return sqlalchemy.types.BigInteger()       

    def convertTypeInt64(self, rtype):
        return sqlalchemy.types.BigInteger()        
    
    def convertTypeInt16(self, rtype):
        return sqlalchemy.types.SmallInteger()        
   
    def convertTypeBool(self, rtype):
        return sqlalchemy.types.Boolean()
    
    def convertTypeReal64(self, rtype):
        return sqlalchemy.types.Float()
    
    def convertTypePickle(self, rtype):
        return sqlalchemy.types.PickleType()

    def convertTypeString(self, rtype):
        if(rtype.dims and rtype.dims[0].shape > 0):
            return sqlalchemy.types.Text(length = rtype.dims[0].shape)
        else:
            return sqlalchemy.types.Text()#}}}
tosa_typemapper = TypeMapperToSQLAlchemy()


class TypeMapperFromSQLAlchemy(VisitorFactory(prefixes=("convert",), #{{{
                                      flags=NF_ELSE)):
    
    def convertAny(self, dbtype, column, typestr="any"):
        if(column.nullable):
            typestr += "?"
        return rtypes.createType(typestr)
    convertNullType = convertAny
    
    def convertBigInteger(self, dbtype, column):
        return self.convertAny(dbtype, column, "int64")
   
    def convertInteger(self, dbtype, column):
        return self.convertAny(dbtype, column, "int32")
    
    def convertSmallInteger(self, dbtype, column):
        return self.convertAny(dbtype, column, "int16")
    
    def convertBoolean(self, dbtype, column):
        return self.convertAny(dbtype, column, "bool")
    
    def convertFloat(self, dbtype, column):
        if(dbtype.precision == 24):
            return self.convertAny(dbtype, column, "real32")
        elif(dbtype.precision == 53):
            return self.convertAny(dbtype, column, "real64")
        else:
            return self.convertAny(dbtype, column, "real64")
            #raise RuntimeError, "Unknown precision: " + str(dbtype)
    
    def convertNumeric(self, dbtype, column):
        return self.convertAny(dbtype, column, "number")
    
    def convertDateTime(self, dbtype, column):
        return self.convertAny(dbtype, column)
    
    def convertTime(self, dbtype, column):
        return self.convertAny(dbtype, column)
    
    def convertDate(self, dbtype, column):
        return self.convertAny(dbtype, column)
    
    def convertInterval(self, dbtype, column):
        return self.convertAny(dbtype, column)

    def convertUnicode(self, dbtype, column):
        return self.convertString(dbtype, column, "string")

    def convertBinary(self, dbtype, column):
        return self.convertString(dbtype, column)

    def convertString(self, dbtype, column, deftype="bytes"):
        if(dbtype.length is None or dbtype.length > 32):
            typestr = "bytes"
        else:
            typestr = "bytes[%d]" % dbtype.length
        if(column.nullable):
            typestr += "?"
        
        return rtypes.createType(typestr)
    
    def convertelse(self,dbtype,column, deftype="int64"):
        if isinstance(dbtype, sqlalchemy.dialects.postgresql.base.ARRAY):
            return self.convertPGArray(dbtype, column)
        else:
            warning('Cannot convert database type: %s for column: %s',repr(dbtype), str(column))
            return self.convertAny(dbtype, column)

    def convertPGArray(self, dbtype, column):
        subtype = self.convert(dbtype.item_type, column)
        subtype.has_missing = False
        name = util.valid_name(column.name)
        typestr = ("[%s:~]<" % name) + str(subtype)
        typestr = ("[%s:~]<" % name) + str(subtype)
        if(column.nullable):
            typestr += "?"
        #JUST USE unknown: postgresql does not enforce array dimensions, so this messes up our code            
        #only thing we can do is make it possible to handle 1 dimension, that should take care of most
        #cases
        return rtypes.createType('unknown')#}}}

sa_typemapper = TypeMapperFromSQLAlchemy()

postgres_types = {}
def _getPostgresTypes(engine):#{{{
    """Returns a list of dictionaries with 
        information on each type in the database"""
    if(not engine in postgres_types):
        res = engine.execute("SELECT oid,typname,typlen,typelem,\
                                typdefault,typbasetype,typnotnull,typtype,typndims \
                        FROM pg_type");
        type_info_lists = res.fetchall()
        type_info_dicts = dict([(til[0],{
                            'name': til[1], 
                            'length':til[2], 
                            'array_element_type_index':til[3], \
                            'default_expression':til[4],
                            'base_type':til[5],\
                            'not_null':til[6], \
                            'type':til[7],
                            'ndims':til[8]
                            }
                             ) for til in type_info_lists])
        postgres_types[engine] = type_info_dicts
    return postgres_types[engine]#}}}
pgtoibidas = {'bool':'bool','oid':'uint32', 'int2':'int16','varchar':'string[]','int4':'int32','text':'string[]', 'float4':'real32', 'float8':'real64'}


mysql_types = {}
def _getMySQLTypes(engine):
    """Returns a list of dictionaries with 
        information on each type in the database"""
    return mysqlid_toibidas
mysqlid_toibidas = {0:'real64',1:'int8',2:'int16',3:'int32',4:'real32',5:'real64',6:'any',7:'any',8:'int64',9:'int32',10:'any',
                    11:'any', 12:'any',13:'uint16',14:'any',15:'string[]',16:'bool',246:'any',247:'any',248:'any',249:'any',250:'any',
                    251:'any',252:'any',253:'string[]',254:'string[]',255:'any'};

def convert(col_descriptor, type, engine, tablename=None):#{{{
    if(type == 'postgres' or type == 'postgresql'):
        fieldnames, sqlnames, subtypes = convert_postgres(col_descriptor, engine)
    elif(type == 'sqlalchemy'):
        fieldnames, sqlnames, subtypes = convert_sqlalchemy(col_descriptor, engine)
    elif(type == 'mysql'):
        fieldnames, sqlnames, subtypes = convert_mysql(col_descriptor, engine)
    else:
        raise RuntimeError, "Unimplemented dialect"
    
    
    table_type = rtypes.TypeTuple(has_missing=False, 
                                    subtypes=tuple(subtypes), 
                                    fieldnames = tuple(fieldnames))

    if(tablename):
        ndims = dimpaths.DimPath(dimensions.Dim("*", name=str(tablename).lower()))
    else:
        ndims = dimpaths.DimPath(dimensions.Dim("*"))

    table_type = rtypes.TypeArray(dims=ndims,
                                    subtypes=(table_type,))
    return (fieldnames, sqlnames, table_type)#}}}

def convert_mysql(col_descriptor, engine):#{{{
    fieldnames = []
    sqlnames = []
    subtypes = []
    for col in col_descriptor:
        name, type_code, display_size, internal_size, precision, scale, null_ok = col
        r = _getMySQLTypes(engine)
        d = r[type_code]
        d = rtypes.createType(d)            
        if(null_ok):
            d = d.setHasMissing(True)
        subtypes.append(d)
        fieldnames.append(util.valid_name(name))
        sqlnames.append(name)
    return (fieldnames, sqlnames, subtypes)#}}}

def convert_postgres(col_descriptor, engine):#{{{
    fieldnames = []
    sqlnames = []
    subtypes = []
    for col in col_descriptor:
        name, type_code, display_size, internal_size, precision, scale, null_ok = col
        r = _getPostgresTypes(engine)
        n = r[type_code]
        if(n['name'] in pgtoibidas):
            d = pgtoibidas[n['name']]
        elif(n['array_element_type_index'] > 0):
            subtype_id = n['array_element_type_index']
            sn = r[subtype_id]
            if sn['name'] == 'char' or sn['name'] == 'text': #string
                sd = 'bytes'
                if(n['length'] == -1):
                    d = sd + "[~]"
                else:
                    d = sd + "[" + str(n['length']) + "]"
                if(not sn['not_null']):
                    sd += "?"
            else:                
                if(sn['name'] in pgtoibidas):
                    sd = pgtoibidas[sn['name']]
                    #if(not sn['not_null']):
                    #    sd += "?"
                    if(n['length'] == -1):
                        d = "[~]<"  
                    else:
                        d = "[" + str(n['length']) + "]<"
                    #JUST USE unknown: postgresql does not enforce array dimensions, so this messes up our code            
                    #only thing we can do is make it possible to handle 1 dimension, that should take care of most
                    #cases
                    d = 'unknown'
                else:
                    d = "unknown"
        else:
            d = "unknown"
        d = rtypes.createType(d)            
        if(null_ok):
            d = d.setHasMissing(True)
        subtypes.append(d)
        fieldnames.append(util.valid_name(name))
        sqlnames.append(name)
    return (fieldnames, sqlnames, subtypes)#}}}

def convert_sqlalchemy(col_descriptor, engine):
    fieldnames = []
    sqlnames = []
    subtypes = []
    for column in col_descriptor:
        fieldnames.append(util.valid_name(column.name))
        sqlnames.append(column.name)
        subtypes.append(sa_typemapper.convert(column.type, column))

    return fieldnames, sqlnames, subtypes

def open_db(*args, **kwargs):
    con = sqlalchemy.create_engine(*args, **kwargs)
    return Connection(con)

class Connection(object):
    def __init__(self, engine, schemaname=None):#{{{
        self._engine = engine
        self._meta = sqlalchemy.MetaData()
        self._schemas = []
        if(schemaname):
            self._meta.reflect(bind=self._engine, schema=schemaname)
        else:
            self._meta.reflect(bind=self._engine)
        
        tables = self._meta.tables.copy()
        self._sa_tables = tables

        info_tables = [table for table in tables.values() if "__info__" in table.name]
        tabledict = {}
        
        for info_table in info_tables:
            name = info_table.name.split('__info__')[0]
            del tables[info_table.name]

            subtables = []
            for tablename,table in list(tables.iteritems()):
                if(tablename == name or '__dim__' in tablename and tablename.split('__dim__')[0] == name):
                    subtables.append(TableRepresentor(self._engine, table))
                    del tables[tablename]
            info_table = TableRepresentor(self._engine, info_table)
            t = CombineRepresentor(self._engine, info_table, subtables)    
            tabledict[name] = t
            
        for table in tables.values():
            t = TableRepresentor(self._engine, table)
            tabledict[table.name] = t
        self._tabledict = tabledict 

        if(not schemaname):
            try:
                schemanames = self._engine.execute('SELECT schema_name FROM information_schema.schemata').fetchall()
            except Exception, e:
                schemanames = []
            
            for sname in schemanames:
                sname = sname[0]
                self.__dict__[sname] = Connection(self._engine, sname);
                self._schemas.append(sname)#}}}

    def __getattr__(self, name):
        return self._tabledict[name]//name

    def _getAttributeNames(self):
        return self._tabledict.keys()
   
    
    def Query(self, query):
        return QueryRepresentor(self._engine, query)

    def __repr__(self):
        res = "Database: " + str(self._engine.url) + "\n"
        tablenames = self._tabledict.keys()
        tablenames.sort()
        if tablenames:
            res += "Tables: " + ", ".join([str(tname) for tname in tablenames])
        if self._schemas:
            res += "Schemas: " + ", ".join(self._schemas)
        return res

    def Store(self, name, rep):#{{{
        if(not name in self._tabledict):
            columns = defaultdict(list)
            rowinfo = []
            rep = rep.Copy()
            for pos, slice in enumerate(rep._slices):
                packdepth = max(len(slice.dims) - 1,0)
                if(packdepth):
                    slice = ops.PackArrayOp(slice,len(slice.dims) - 1)
                
                if(slice.dims):
                    stype = tosa_typemapper.convert(slice.type) 
                    if(stype is False):
                        pickled = True
                        stype = sqlalchemy.types.Text()
                    else:
                        pickled = False
                    nullable = slice.type.has_missing
                    columns[slice.dims[0]].append(sqlalchemy.Column(slice.name, stype, nullable = nullable))
                    dimname = str(slice.dims[0].name)
                    rowinfo.append({'spos':pos, 'name':slice.name, 'pickle':pickled, 'type':str(slice.type), 'packdepth': packdepth, 'dimname':dimname,'val':""})
                else:
                    dimname = Missing
                    pickled = True
                    val = rep.Get(slice.name).Cast("pickle").Each(buffer,dtype="pickle").ToPython()
                    rowinfo.append({'spos':pos, 'name':slice.name, 'pickle':pickled, 'type':str(slice.type), 'packdepth': packdepth, 'dimname':dimname, 'val':val})
            
            tables = {}
            for key, value in columns.iteritems():
                if(len(columns) > 1):
                   tablename = name + "__dim__" + key.name
                else:
                   tablename = name
                newtable = sqlalchemy.Table(tablename, self._meta, *value) 
                newtable.create(bind=self._engine, checkfirst=True)
                tables[key] = newtable
            
            if(len(columns) > 1 or any([row['pickle'] or row['dimname'] is None for row in rowinfo])):
                self.Store(name + "__info__",python.Rep(rowinfo, dtype="[info:*]<{spos=int,name=bytes,pickle=bool,type=bytes,dimname=bytes?,packdepth=int,val=bytes}"))
               
                tablelist = []
                for t in tables.values():
                    self._sa_tables[t.name] = t
                    table = self._meta.tables[t.name]
                    t = TableRepresentor(self._engine, table)
                    tablelist.append(t)
                infotable = self._meta.tables[name + "__info__"]
                del self._tabledict[name + "__info__"]
                info = TableRepresentor(self._engine, infotable)
                t = CombineRepresentor(self._engine, info, tablelist)
            else:
                self._sa_tables[name] = tables.values()[0]
                table = self._meta.tables[name]
                t = TableRepresentor(self._engine, table)
            self._tabledict[name] = t

        self._append(name, rep)
        return self._tabledict[name]#}}}
    
    def _append(self, name, rep):#{{{
        tbl = self._tabledict[name]
        if(isinstance(tbl,CombineRepresentor)):
            rep = rep.Array(tolevel=1) 
            pickle = tbl._info[tbl._info.pickle == True].name()
            rep = rep.To(*pickle, Do=_.Cast("pickle").Each(buffer,dtype="pickle"))
            rep = rep.Copy()
            for col in tbl._info.Dict().ToPython():
                if 'dimname' in col:
                    continue
                val = rep.Get(col['name']).ToPython()
                assert col['val'] == val, "Scalar value field: " + col['name']  + " does not match."
            for table in tbl._tablelist:
                self._append_to_table(table,rep)
        elif(isinstance(tbl,TableRepresentor)):
            self._append_to_table(tbl, rep)
        else:
            raise RuntimeError, "Cannot add data to a query"
            

    def _append_to_table(self, table, rep):
        w = dict([(s.name,s) for s in rep._slices])
        cols = [slice for slice in table._slices if slice.name in w]
        
        for s in cols:
            if not s.type >=w[s.name].type:
                raise RuntimeError, "Incompatible types: " + str(s.type) + " " + str(w[s.name].type)
        rep = rep.Get(*[slice.name for slice in cols])
        
        trep = rep.Tuple()
        tbdims = set([slice.dims for slice in trep._slices])
        assert len(tbdims) == 1 and len(tbdims.pop()) == 1, \
                "Only a representor object with one common parent dimension can be \
                    stored as a table"

        values = rep.Dict(with_missing=True).ToPython()
        self._engine.execute(self._sa_tables[table._tablename].insert(), values)#}}}

    def Drop(self, name):#{{{
        tbl = self._tabledict[name]
        if isinstance(tbl, CombineRepresentor):
            info_table = tbl._info
            self._sa_tables[info_table._tablename].drop(bind=self._engine, checkfirst=True)
            del self._sa_tables[info_table._tablename]

            tablelist = tbl._tablelist
            for tbl in tablelist:
                self._sa_tables[tbl._tablename].drop(bind=self._engine, checkfirst=True)
                del self._sa_tables[tbl._tablename]
        else:
            self._sa_tables[name].drop(bind=self._engine, checkfirst=True)
            del self._sa_tables[name]
        del self._tabledict[name]#}}}

class CombineRepresentor(wrapper.SourceRepresentor):#{{{
    def __init__(self, engine, info, tablelist):
        self._tablelist = tablelist
        self._info = info
        dims = set()
        nslicelist = []
        tablepos = {}
        for pos, table in enumerate(tablelist):
            nslicelist.append(dict([(slice.name,slice) for slice in table._slices]))
            if('__dim__' in table._tablename):
                tablepos[table._tablename.split('__dim__')[1]] = pos
            else:
                tablepos[table._tablename] = pos
        
        nslices = []
        for row in info.Sort(info.spos).Dict()():
            if not 'dimname' in row:
                dimpos = 0
            else:
                dimpos = 1

            if row['pickle']:
                rtype = rtypes.createType("pickle", dimpos)
                ctype = rtypes.createType(row['type'], dimpos)
            else:
                rtype = rtypes.createType(row['type'], dimpos)

            if not 'dimname' in row:
                if not 'val' in row:
                    continue
                res_slice = ops.DataOp(row['val'], name=row['name'], rtype=rtype)
            elif(len(tablelist) > 1):
                assert row['dimname'] in tablepos, "Cannot find table for dim : " + row['dimname']
                tslices = nslicelist[tablepos[row['dimname']]]
                res_slice = tslices[row['name']]
            else:
                tslices = nslicelist[0]
                res_slice = tslices[row['name']]
               
            if not res_slice.type == rtype:
                res_slice = ops.CastOp(res_slice,rtype)

            if row['pickle']:
                res_slice = ops.CastOp(res_slice,ctype)
            if row['packdepth']:
                res_slice = ops.UnpackArrayOp(res_slice, row['packdepth'])
            nslices.append(res_slice)
        self._initialize(tuple(nslices))#}}}

class TableRepresentor(wrapper.SourceRepresentor):#{{{
    def __init__(self, engine, table):
        self._tablename = table.name
        fieldnames, sqlnames, table_type = convert(table.columns,'sqlalchemy',engine, table.name)
        nslices = repops_slice.UnpackTuple._apply(ops.UnpackArrayOp(SQLOp(engine, table, sqlnames, table_type, util.valid_name(table.name))))
        nnslices = []
        for slice in nslices:
            if slice.type.hasMissing():
                slice = ops.NoneToMissingOp(slice)
            nnslices.append(slice)
                
        self._initialize(tuple(nnslices))#}}}

class QueryRepresentor(wrapper.SourceRepresentor):#{{{
    def __init__(self, engine, query):
        res = engine.execute('select * from (%s) table_alias_33 limit 0;' %query)
        try:
            fieldnames, sqlnames, query_type = convert(res.context.compiled.statement.columns, "sqlalchemy", engine)
        except AttributeError:
            fieldnames, sqlnames, query_type = convert(res.cursor.description, res.dialect.name, engine)
        res.close()            
        

        if not query.lower().startswith('select'):
            error('Query should start with SELECT')
        query = query[7:] #drop select, as it is added again by sql.expression.select
        query = sqlalchemy.sql.expression.select([sqlalchemy.sql.expression.text(query)])
        nslices = repops_slice.UnpackTuple._apply(ops.UnpackArrayOp(SQLOp(engine, query, sqlnames,  query_type, "query_result")))
        nnslices = []
        for slice in nslices:
            if slice.type.has_missing:
                slice = ops.NoneToMissingOp(slice)
            nnslices.append(slice)
        self._initialize(tuple(nnslices))#}}}

class SQLResultEdge(query_graph.Edge):#{{{
    __slots__ = ["realedge","pos"]
    def __init__(self, source, target, realedge, fieldpos):
        self.realedge = realedge
        self.pos = fieldpos
        query_graph.Edge.__init__(self, source, target)#}}}

class SQLTransEdge(SQLResultEdge):#{{{
    __slots__ = ["targetpos"]
    def __init__(self, source, target, realedge, fieldpos, tpos):
        self.targetpos = tpos
        SQLResultEdge.__init__(self,source,target,realedge,fieldpos)#}}}
    

class SQLPlanner(VisitorFactory(prefixes=("eat","expressionEat"), 
                                      flags=NF_ELSE), manager.Pass):
    after=set([create_graph.CreateGraph, annotate_replinks.AnnotateRepLinks])
    before=set([serialize_exec.SerializeExec])

    @classmethod
    def run(cls, query, run_manager, debug_mode=False):#{{{
        self = cls()
        self.graph = run_manager.pass_results[create_graph.CreateGraph]
        self.expressions = run_manager.pass_results[annotate_replinks.AnnotateRepLinks]

        self.done = set()
        self.sql_obj = set()
        self.eat_front = self.graph.getSources()

        while(self.eat_front):
            self.next_round = set()
            if(debug_mode):
                from ..passes.cytoscape_vis import DebugVisualizer
                DebugVisualizer.run(query,run_manager)
            for node in list(self.eat_front):
                self.eat(node)
            
            self.eat_front = self.next_round

        for node in self.sql_obj:
            if(isinstance(node, Query)):
                self.finish(node)
            else:
                self.graph.dropNode(node)#}}}

    #UTILS
    def allSQLParams(self, in_slice, expression):#{{{
        edges = self.graph.edge_target[in_slice]
        paramedges = set()
        sqlparamedges = set()
        for edge in edges:
            if(edge.source in expression.all_slices):
                continue
            if isinstance(edge, query_graph.ParamEdge):
                paramedges.add(edge)
            elif isinstance(edge, SQLResultEdge):
                sqlparamedges.add(edge.realedge)
        return paramedges == sqlparamedges#}}}

    def getSQLEdge(self, edge):#{{{
        etargets = self.graph.edge_target[edge.target]
        for etarget in etargets:
            if isinstance(etarget,SQLResultEdge):
                if(etarget.realedge == edge):
                   return etarget
        return False#}}}

    def compatible(self,elem1,elem2):#{{{
        if(isinstance(elem1, SQLValue) or isinstance(elem2, SQLValue)):
            return True

        if(isinstance(elem2, SQLTerm)):
            elem1, elem2 = elem2, elem1

        if(isinstance(elem1, SQLTerm)):
            s = elem1.getSource(self.graph)
            if(isinstance(elem2, SQLQuery)):
                return s == elem2 or s is None
            elif(isinstance(elem2, SQLTerm)):
                s2 = elem2.getSource(self.graph)
                return s == s2 or s is None or s2 is None
        elif(isinstance(elem1, SQLQuery)):
            if(isinstance(elem2, SQLQuery)):
                return elem1 == elem2        
         
        raise RuntimeError, "Encountered unknown type of SQL object: " + str(elem1) + " and " + str(elem2)#}}}
            
    def addPairToGraph(self, node, node2):#{{{
        self.graph.addNode(node)
        self.graph.addNode(node2)
        self.graph.addEdge(query_graph.ParamEdge(node,node2,"slice"))#}}}

    #SOURCE OBJECTS
    def eatSQLOp(self, node):#{{{
        if(isinstance(node.query, sqlalchemy.sql.expression.Executable)):
            o = Query(node.conn, TableQuery(node.query, node.columns))
        else:
            o = Query(node.conn, Table(node.query, node.columns))
        
        self.graph.addNode(o)
        self.next_round.add(o)
        self.sql_obj.add(o)

        etargets = list(self.graph.edge_source[node])
        assert len(etargets) == 1, "Expected one target slice after sql op"
        anode = etargets[0].target
        assert isinstance(anode,ops.UnpackArrayOp), "Expected array unpack operation after sql op"
        
        targets = [edge.target for edge in self.graph.edge_source[anode]]
        columns = o.getColumns()
        ncolumns = []
        
        for pos,target in enumerate(targets):
            assert isinstance(target,ops.UnpackTupleOp), "Expected tuple unpack operation after sql op array unpack"
            ncolumns.append(columns[target.tuple_idx])
            for edge in self.graph.edge_source[target]:
                self.graph.addEdge(SQLResultEdge(o, edge.target, edge, pos))

        o.setColumns(ncolumns)
        #}}}
    
    def eatOp(self, node):#{{{
        pass#}}}

    def eatDataOp(self, node):#{{{
        o = Value(node.data)
        self.graph.addNode(o)
        self.next_round.add(o)
        self.sql_obj.add(o)

        for edge in self.graph.edge_source[node]:
            self.graph.addEdge(SQLResultEdge(o, edge.target, edge, 0))#}}}
    
    #SQL QUERY
    def eatElement(self, node):#{{{
        self.expressionEat(node)#}}}


    def eatQuery(self, node):#{{{
        for edge in list(self.graph.edge_source[node]):
            if(isinstance(edge.target, ops.NoneToMissingOp)):
                self.graph.dropEdge(edge)
                for tedge in list(self.graph.edge_source[edge.target]):
                    self.graph.addEdge(SQLResultEdge(node, tedge.target, tedge, edge.pos))
        self.expressionEat(node)#}}}
                

    #SQL VALUE
    def eatValue(self, node):#{{{
        return self.expressionEat(node)#}}}

    #EXPRESSIONS
    def expressionEatElement(self, node):#{{{
        targets = [edge.target for edge in self.graph.edge_source[node]]
        links = self.graph.node_attributes['links']
        expressions = set([links[target] for target in targets if target in links])

        for expression in expressions:
            if expression in self.done:
                continue
            if all([self.allSQLParams(in_slice, expression) for in_slice in expression.in_slices]):
                self.expressionEat(expression)#}}}

    def expressionEatExpression(self, node):#{{{
        pass#}}}
        
    def expressionEatMatchExpression(self, expression):#{{{
        used_sources = expression.getUsedSources()
        
        cledge,credge = expression.getComparisonEdges(self.graph)
        scledge = self.getSQLEdge(cledge)
        scredge = self.getSQLEdge(credge)
        clsource = scledge.source
        crsource = scredge.source

        if used_sources[0]:
            dimpathset = expression.getDim(0, self.graph)
            if(len(dimpathset) > 1 or len(dimpathset.pop()) > 1):
                return
        
            ledges,olnodes = expression.getInfo(0, self.graph)
            sledges = [self.getSQLEdge(ledge) for ledge in ledges]
            lsources = list(set([sledge.source for sledge in sledges]))
            if(len(lsources) > 1):
                return
            lsource = lsources[0]
        else:
            if isinstance(clsource, Query):
                lsource = clsource
            elif(isinstance(clsource, Column) and clsource.ref_query):
                lsource = clsource.ref_query
            else:
                return
                

        if used_sources[1]:
            dimpathset = expression.getDim(1, self.graph)
            if(len(dimpathset) > 1 or len(dimpathset.pop()) > 1):
                return
            redges,ornodes = expression.getInfo(1, self.graph)
            sredges = [self.getSQLEdge(redge) for redge in redges]
            rsources = list(set([sredge.source for sredge in sredges]))
            if(len(rsources) > 1):
                return
            rsource = rsources[0]
        else:
            if isinstance(crsource, Query):
                rsource = crsource
            elif(isnstance(crsource, Column) and crsource.ref_query):
                rsource = crsource.ref_query
            else:
                return

        if(not lsource.conn == rsource.conn):
            return
        
        jointype = expression.getType()
        if not jointype in set(["inner","left","right"]):
            return 

        lcol = clsource.getColumn(scledge.pos)
        rcol = crsource.getColumn(scredge.pos)
        cond = Term(lambda x,y: x == y,lcol,rcol)
        nqueryobj = lsource.join(rsource, cond, jointype)

        nlcolumns = len(lsource.columns)
        self.graph.addNode(nqueryobj)

        if used_sources[0]:
            for onode, ledge in zip(olnodes,sledges):
                for edge in self.graph.edge_source[onode]:
                    self.graph.addEdge(SQLResultEdge(nqueryobj,edge.target,edge, ledge.pos))
        if used_sources[1]: 
            for onode, redge in zip(ornodes,sredges):
                for edge in self.graph.edge_source[onode]:
                    self.graph.addEdge(SQLResultEdge(nqueryobj,edge.target,edge, nlcolumns + redge.pos))

        self.next_round.add(nqueryobj)
        self.sql_obj.add(nqueryobj)
        self.done.add(expression)#}}}

    def expressionEatFilterExpression(self, expression):#{{{
        #check filter dim
        dimpathset = expression.getDims()
        if(len(dimpathset) > 1 or len(dimpathset.pop()) != 0):
            return

        fedges,cedge, onodes = expression.getInfo(self.graph)
        
        if(not isinstance(cedge.source.type,(rtypes.TypeBool,rtypes.TypeSlice))):
            return

        scedge = self.getSQLEdge(cedge) 
        sfedges = [self.getSQLEdge(fedge) for fedge in fedges]

        fsources = list(set([sfedge.source for sfedge in sfedges]))
        if(len(fsources) != 1 or not isinstance(fsources[0],Query)):
            return 

        fsource = fsources[0]
        csource = scedge.source

        nsource = fsource.copy()
        #check filter type
        if(isinstance(cedge.source.type,rtypes.TypeBool)):
            if(not isinstance(csource, Term)):
                return
            nsource = nsource.addWhere(csource)

        else: #typeslice
            if(not isinstance(csource, Value)):
                return
            if not isinstance(csource.data,slice):
                return
            if not csource.data.step is None: 
                return
            start = csource.data.start
            stop = csource.data.stop
            if(not start is None and start < 0):
                return
            if(not stop is None and stop < 0):
                return
            
            nsource = nsource.addLimit(start, stop)
        

        self.graph.addNode(nsource)
        
        for onode, fedge in zip(onodes,sfedges):
            for edge in self.graph.edge_source[onode]:
                self.graph.addEdge(SQLResultEdge(nsource,edge.target,edge, fedge.pos))
        
        self.next_round.add(nsource)
        self.sql_obj.add(nsource)
        self.done.add(expression)#}}}
    
    def expressionEatBinFuncElemExpression(self, expression):#{{{
        sig, op = expression.getOp()
        #check sig
        #check sql compatibility
        if(not op in SQLOperators and not (sig + op) in SQLOperators):
            return

        if(op in SQLOperators):
            opfunc = SQLOperators[op]
        else:
            opfunc = SQLOperators[sig + op]

        #check dimensions
        leftedge,rightedge = expression.getLeftRightInEdges(self.graph)
        ldims, rdims = leftedge.source.dims, rightedge.source.dims
        if len(ldims) > 1 or len(rdims) > 1:
            return
        if ldims != rdims and len(rdims) == len(ldims):
            return
        
        sqlleftedge = self.getSQLEdge(leftedge)
        sqlrightedge = self.getSQLEdge(rightedge)

        left = sqlleftedge.source
        right = sqlrightedge.source

        out = expression.getOutSlice()
        res = Term(opfunc, left.getColumn(sqlleftedge.pos), right.getColumn(sqlrightedge.pos))
        self.graph.addNode(res)
        self.graph.addEdge(SQLTransEdge(left, res, sqlleftedge, sqlleftedge.pos, 0))
        self.graph.addEdge(SQLTransEdge(right, res, sqlrightedge, sqlrightedge.pos, 1))

        out_edges = expression.getOutEdges(self.graph)
        for out_edge in out_edges:
            self.graph.addEdge(SQLResultEdge(res,out_edge.target,out_edge,0))

        self.next_round.add(res)
        self.sql_obj.add(res)
        self.done.add(expression)#}}}


    def finish(self, node):#{{{
        #filter out unnecessary fields
        used_pos = set()
        posresult = dict()
        results = []
        for edge in list(self.graph.edge_source[node]):
            if(isinstance(edge.realedge,query_graph.ParamEdge)):
                used_pos.add(edge.pos)
                posresult[edge.pos]= edge.realedge.source

        used_pos = list(used_pos)
        opos_to_npos = dict([(opos,npos) for npos, opos in enumerate(used_pos)])
        
        columns = [node.columns[pos] for pos in used_pos]
        results = [posresult[pos] for pos in used_pos]
        fnode = node.copy()
        fnode.setColumns(columns)

        #compile, create unpack slices
        packdepth = [0] * len(results)
        for i in range(len(results)):
            packdepth[i] = len(results[i].dims) - 1
            if(len(results[i].dims) < 1):
                results[i] = ops.UnpackArrayOp(results[i])
            elif(len(results[i].dims) > 1):
                results[i] = ops.PackArrayOp(results[i], len(results[i].dims) - 1)

        tslice = ops.PackTupleOp(results)
        aslice = ops.PackArrayOp(tslice)
        
        sslice = SQLOp(fnode.conn, fnode.compile(), [column.name for column in columns], aslice.type)
        rslice = ops.UnpackArrayOp(sslice)
        self.addPairToGraph(sslice,rslice)
        
        rslices = []
        for oslice, i in zip(results, range(len(results))):
            r = ops.UnpackTupleOp(rslice,i)
            self.addPairToGraph(rslice,r)

            if(r.bookmarks):
                r2 = ops.ChangeBookmarkOp(r)
                r2.bookmarks = oslice.bookmarks
                self.addPairToGraph(r,r2)
                r= r2

            if(r.type.has_missing):
                r2 = ops.NoneToMissingOp(r)
                self.addPairToGraph(r,r2)
                r= r2

            if(packdepth[i] > 0):
                r2 = ops.UnpackArrayOp(r, packdepth[i])
                self.addPairToGraph(r,r2)
                r= r2
            elif(packdepth[i] < 0): 
                r2 = ops.PackArrayOp(r)
                self.addPairToGraph(r,r2)
                r= r2

            rslices.append(r)
      
        for edge in list(self.graph.edge_source[node]):
            self.graph.dropEdge(edge.realedge)
            self.graph.dropEdge(edge)
            
            redge = edge.realedge
            if(isinstance(redge,query_graph.ParamEdge)):
                redge.source = rslices[opos_to_npos[edge.pos]]
                self.graph.addEdge(redge)#}}}
        
        
SQLOperators = {'Equal':lambda x, y: x==y,
                'NotEqual': lambda x, y: x!=y,
                'Less': lambda x, y: x < y,
                'LessEqual': lambda x, y: x <= y,
                'Greater': lambda x,y: x > y,
                'GreaterEqual': lambda x,y: x >= y,
                'boolboolAnd': sql.and_,
                'boolboolOr': sql.or_,
                'boolboolXor': lambda x, y: x ^ y,
                'simple_arithAdd': lambda x, y: x + y,
                'string_add_stringAdd': lambda x, y: x + y,
                'Subtract': lambda x, y: x - y,
                'Multiply': lambda x, y: x * y,
                'Modulo': lambda x, y: x % y,
                "Divide": lambda x, y: x / y,
                'Power': lambda x, y: x ** y
                }


class SQLOp(ops.ExtendOp):#{{{
    __slots__ = []
    passes = [SQLPlanner]
    def __init__(self, conn, query, columns, rtype, name="result"):
        self.conn = conn
        self.query = query
        self.columns = columns
        ops.ExtendOp.__init__(self,name=name,rtype=rtype)

    def py_exec(self):
        util.debug('Executing query: %s', str(self))
        if(isinstance(self.query, sqlalchemy.sql.expression.Executable)):
            res = self.conn.execute(self.query)
        else:
            res = self.conn.execute(sql.select([self.query]))
        result = res.fetchall()
        ndata = nested_array.NestedArray(result,self.type)
        return python.ResultOp.from_slice(ndata,self)

    def __str__(self):
        if(isinstance(self.query, sqlalchemy.sql.expression.Executable)):
            return ops.ExtendOp.__str__(self) + ":\n" +  str(self.query.compile())
        else:
            return ops.ExtendOp.__str__(self) + ":\n" +  str(sql.select([self.query]).compile())#}}}

class Element(query_graph.Node):
    pass

class Column(Element):#{{{
    def __init__(self, table, name, id):
        self.table = table
        self.name = name
        self.id = id
        self.ref_query = None

    def getColumn(self, pos):
        assert pos == 0, "Cannot request column past position 0 for Column object"
        return self

    def realias(self, realias_dict):
        if(self.table in realias_dict):
            return Column(realias_dict[self.table],self.name, self.id)
        else:
            return self

    def recol(self, column_dict):
        return column_dict.get(self,self)

    def compile(self):
        #pos = self.table.getColPos(self.id)
        return sqlalchemy.sql.expression.literal_column(self.table.getAlias() + '.' + self.name)
        #return list(self.table.compile().columns)[pos]

    def getTables(self):
        return set([self.table])

    def __str__(self):
        return str(self.table) + ":" + str(self.name)#}}}

class Value(Column):#{{{
    def __init__(self, data):
        self.data = data
        self.ref_query = None

    def compile(self):
        return self.data

    def realias(self, realias_dict):
        return self

    def recol(self, realias_dict):
        return self

    def getTables(self):
        return set()

    def __str__(self):
        return str(self.data)#}}}

class Term(Column):#{{{
    def __init__(self, func, *sources):
        self.sources = sources
        rquery = set([source.ref_query for source in sources])
        if(len(rquery) == 1):
            self.ref_query = rquery.pop()
        else:
            self.ref_query = None
        self.func = func

    def realias(self, realias_dict):
        nsources = [source.realias(realias_dict) for source in self.sources]
        return Term(self.func, *nsources)
        
    def recol(self, column_dict):
        nsources = [source.recol(column_dict) for source in self.sources]
        return Term(self.func, *nsources)

    def compile(self):
        nc = [source.compile() for source in self.sources]
        return self.func(*nc)

    def getTables(self):
        return reduce(operator.__or__,[source.getTables() for source in self.sources],set())

    def __str__(self):
        return str(self.compile())#}}}

class Table(object):#{{{
    def __init__(self, source_descriptor, column_names):
        self.source_descriptor = source_descriptor
        self.colids = range(len(list(column_names)))
        self.column_names = column_names
        self._alias = self.getName() + str(alias_name_id.next()) 

    def getSource(self):
        return self.source_descriptor

    def getColumns(self):
        cols = self.column_names
        return [Column(self, col, colid) for col,colid in zip(cols,self.colids)]

    def getColPos(self, id):
        return self.colids.index(id)

    def realias(self, realias_dict):
        return realias_dict.get(self, self)

    def alias(self, realias_dict):
        n = AliasTable(self)
        realias_dict[self] = n
        return n
    
    def getTables(self):
        return set([self])

    def compile(self):
        return self.source_descriptor.alias(self._alias)

    def getName(self):
        return self.source_descriptor.name

    def getAlias(self):
        return self._alias

    def __str__(self):
        return str(self.compile().alias(self._alias))#}}}

class TableQuery(Table):
    def __init__(self, source_descriptor, columns):
        self.alias = 'query' + str(alias_name_id.next()) 
        Table.__init__(self, source_descriptor.alias(self._alias), columns)
    
    def getName(self):
        return self._alias

class AliasSubQuery(TableQuery):#{{{
    def __init__(self, source):
        self.cached = None
        TableQuery(source, source.getColumns())

    def realias(self, realias_dict):
        self = realias_dict[self]
        self.source_descriptor = self.source_descriptor.realias(realias_dict)
        self.cached = None
        return self

    def alias(self, realias_dict):
        n = self.source_descriptor.to_subquery()
        realias_dict[self] = n
        return n

    def getTables(self):
        return set([self]) | self.source_descriptor.getTables()

    def compile(self):
        if(self.cached is None):
            self.cached = self.source_descriptor.compile().alias(self._alias)
        return self.cached#}}}

class AliasTable(Table):#{{{
    def __init__(self, origtable):
        self.origtable = origtable
        self._alias = origtable.getName() + str(alias_name_id.next())
        Table.__init__(self,origtable.getSource().alias(self._alias), origtable.column_names)

    def alias(self, realias_dict):
        return self.origtable.alias(realias_dict)#}}}

    def getName(self):
        return self.origtable.getName()


class Join(Table):#{{{
    def __init__(self, left, right, jointype, condition):
        assert not left.getTables() & right.getTables(), "Overlap in tables in join condition"
        assert not (condition.getTables() - (left.getTables() | right.getTables())), "Condition has tables outside join"
        self.left = left
        self.right = right
        self.jointype = jointype
        self.condition = condition

    def realias(self, realias_dict):
        n = Join(self.left.realias(realias_dict), self.right.realias(realias_dict), self.jointype, self.condition.realias(realias_dict))

    def getTables(self):
        return self.left.getTables() | self.right.getTables() | self.condition.getTables()

    def getColumns(self):
        return self.left.getColumns() + self.right.getColumns()

    def compile(self):
        lsource = self.left.compile()
        rsource = self.right.compile()
        if(self.jointype == "inner"):
            fromobj = lsource.join(rsource, onclause=self.condition.compile())
        elif(self.jointype == "left"):
            fromobj = lsource.join(rsource, onclause=self.condition.compile(),isouter=True)
        elif(self.jointype == "right"):
            fromobj = rsource.join(lsource, onclause=self.condition.compile(),isouter=True)
        return fromobj#}}}


class Query(Element):#{{{
    def __init__(self, conn, from_obj):
        self.conn = conn
        assert isinstance(from_obj, Table), "from obj should be a table"

        self.from_obj = from_obj
        self.columns = self.from_obj.getColumns()
        self.limit = None
        self.offset = None
        self.distinct = False
        self.whereclause = []
        self.groupby_clause = []
        self.havingclause = []
        self.orderedby_clause = []

    def copy(self):
        s =  copy.copy(self)
        s.whereclause = list(s.whereclause)
        s.groupby_clause = list(s.groupby_clause)
        s.havingclause = list(s.havingclause)
        s.orderedby_clause = list(s.orderedby_clause)
        return s

    def hasTables(self, tables):
        return len(set(tables) - self.getTables()) == 0

    def setColumns(self, cols):
        assert not (reduce(operator.__or__,[col.getTables() for col in cols],set()) - self.from_obj.getTables()), "Column tables not in from obj"
        self.columns = cols

    def getColumn(self, pos):
        c = copy.copy(self.columns[pos])
        c.ref_query = self
        return c


    def getColumns(self):
        return self.columns

    def _clause_subquery(self, clause):
        nself = self.to_subquery()
        column_dict = dict(zip(self.columns,nself.columns))
        return (nself, clause.recol(column_dict))

    def addWhere(self, clause):
        if (self.groupby_clause or self.limit or self.offset):
            self, clause = self._clause_subquery(clause)

        assert not clause.getTables() - self.from_obj.getTables(), "Where tables not in from obj"
        self.whereclause.append(clause)
        return self

    def addHaving(self, clause):
        assert self.groupby_clause and not (self.limit or self.offset), "Having clause should be added to queries with group, and without offset/limit"
        assert not clause.getTables() - self.from_obj.getTables(), "Having tables not in from obj"
        self.havingclause.append(clause)
        return self

    def addGroupBy(self, clause):
        if (self.limit or self.offset):
            assert not self.groupby_clause, "Cannot perform group as limit or offset has been set inbetween"
            self, clause = self._clause_subquery(clause)

        assert not clause.getTables() - self.from_obj.getTables(), "Group tables not in from obj"
        self.groupby_clause.append(clause)
        return self

    def addOrderedBy(self, clause):
        if (self.limit or self.offset or self.groupby_clause):
            assert not self.orderedby_clause, "Cannot perform order as limit or offset has been set inbetween"
            self, clause = self._clause_subquery(clause)
 
        assert not clause.getTables() - self.from_obj.getTables(), "Group tables not in from obj"
        self.orderedby_clause.append(clause)
     
    def setDistinct(self, distinct):
        assert not (self.limit or self.offset or self.groupby_clause), "Cannot set distinct on query with group"
        self.distinct = distinct
        return self

    def addLimit(self, start, stop):
        assert start is None or start >= 0, "Start cannot be smaller than 0"
        assert stop is None or stop >= 0, "Stop cannot be smaller than 0"

        if(not start is None):
            xoffset = max(self.offset,0)#None to 0
            if(self.limit is None):
                self.offset = xoffset + start
            else:
                start = min(start, self.limit)
                self.offset = xoffset + start
                self.limit -= start
            
        if(not stop is None):
            if(self.limit is None):
                self.limit = stop - max(start,0)
            else:
                self.limit = min(stop - max(start,0), self.limit)
        return self


    def realias(self, realias_dict):
        n = Query(self.conn, self.from_obj.realias(realias_dict))
        n.columns = [col.realias(realias_dict) for col in self.columns]
        n.whereclause = [wc.realias(realias_dict) for wc in self.whereclause]
        n.groupby_clause = [gc.realias(realias_dict) for gc in self.groupby_clause]
        n.orderedby_clause = [oc.realias(realias_dict) for oc in self.orderedby_clause]
        n.havingclause = [hc.realias(realias_dict) for hc in self.havingclause]
        n.limit = self.limit
        n.offset = self.offset
        n.distinct = self.distinct
        return n

    def getTable(self):
        return self

    def getTables(self):
        return self.from_obj.getTables()

    def to_subquery(self):
        nt = AliasSubQuery(self)
        n = Query(self.conn,nt)
        return n

    def _from_compile(self):
        return self.from_obj.compile()

    def compile(self):
        columns = [c.compile() for c in self.columns]
        from_obj = self._from_compile()
        whereclause = sql.and_(*[wc.compile() for wc in self.whereclause])
        return sql.select(columns,whereclause=whereclause, from_obj = from_obj, limit=self.limit, offset = self.offset, distinct=self.distinct, use_labels = True)

    def join(self, other, condition, jointype):
        if (self.groupby_clause or self.distinct or self.limit or self.offset):
            self, condition = self._clause_subquery(condition)
        if (other.groupby_clause or other.distinct or other.limit or other.offset):
            other, condition = other._clause_subquery(condition)

        tableintersect = self.getTables() & other.getTables()
        if(tableintersect):
            realias_dict = {}
            for table in tableintersect:
                table.alias(realias_dict)
            other = other.realias(realias_dict)
            condition = condition.realias(realias_dict)
        j = Join(self.from_obj, other.from_obj, jointype, condition)
        assert self.conn == other.conn, "Mismatch in connections"
        q = Query(self.conn, j)
        q.setColumns(self.getColumns() + other.getColumns())
        q.whereclause = self.whereclause + other.whereclause
        return q
    def __str__(self):
        return str(self.compile())#}}}






import copy
from collections import defaultdict
from itertools import izip_longest

from constants import *
from utils import util
from itypes import rtypes,dimpaths
from ops import *
from query_graph import Node

_delay_import_(globals(),"itypes","dimensions","typeops","convertors","casts")
_delay_import_(globals(),"itypes.type_attribute_freeze","freeze_protocol")

class Op(Node):
    __slots__ = ["link"]

    def assignLink(self,link):
        if(not hasattr(self, "link")):
            self.link = link

#pylint: disable-msg=E1101
class UnaryOp(Op):#{{{
    __slots__ = []
    """A unary operation is not only a description of an operation, but
    also the description of the output of that operation. As such, each unary
    op has a name, a type and a dimension attribute, as well as bookmarks.
    """

    __slots__ = ['name', 'type', 'dims','bookmarks']


    def __init__(self, name, rtype = rtypes.unknown, dims=dimpaths.DimPath(), bookmarks=set()):
        """Creates a slice object.

        :param name: name of slice (string)
        :param type: type of slice, optional, default = unknown
        :param dims: tuple of Dim objects, optional, default = ()
        """
        assert isinstance(name, basestring), "Name of slice should be a string"
        assert (name.lower() == name), "Name should be in lowercase"
        assert isinstance(rtype, rtypes.TypeUnknown), "Invalid type given"
        assert isinstance(dims, dimpaths.DimPath), "Dimensions of a slice should be a DimPath"
        assert isinstance(bookmarks,set), "Bookmarks should be a set"
        assert all([isinstance(bm, basestring) for bm in bookmarks]),"Bookmarks should be a string"

        self.name = name
        self.type = rtype
        self.dims = dims
        self.bookmarks = bookmarks
    
    def __repr__(self):
        res = self.name
        res += "="
        if(self.dims):
            dimstr = []
            for pos, dim in enumerate(self.dims):
                dimstr.append("[" + str(dim) + "]")
            res += "<".join(dimstr) + "<"
                
        res += str(self.type)
        return res#}}}

class ExtendOp(UnaryOp):#{{{
    __slots__ = []
#}}}

class DataOp(UnaryOp):#{{{
    __slots__ = ['data']
    def __init__(self, data, name=None, rtype=None, dims=dimpaths.DimPath(), bookmarks=set()):
        self.data = data
        UnaryOp.__init__(self, name, rtype, dims, bookmarks)#}}}

class UnaryUnaryOp(UnaryOp):#{{{
    __slots__ = ["source"]
    def __init__(self, source, name=None, rtype=None, dims=None, bookmarks=None):
        if(name is None):
            name = source.name
        if(rtype is None):
            rtype = source.type
        if(dims is None):
            dims = source.dims
        if(bookmarks is None):
            bookmarks = source.bookmarks
        self.source = source
        assert isinstance(source,Op),"Source of UnaryUnaryOpOp should be a slice"
        UnaryOp.__init__(self, name, rtype, dims, bookmarks)#}}}

class ChangeBookmarkOp(UnaryUnaryOp):#{{{
    __slots__ = []

    def __init__(self,source,add_bookmark=None,update_auto_bookmarks=None):
        nbookmarks = source.bookmarks.copy()

        if(not update_auto_bookmarks is None):
            for bm in list(nbookmarks):
                if(bm.startswith('!')):
                    nbookmarks.discard(bm)
                    nbookmarks.add("!" + update_auto_bookmarks + bm[1:])

        if(not add_bookmark is None):
            nbookmarks.add(add_bookmark)
        
        UnaryUnaryOp.__init__(self, source, bookmarks=nbookmarks)#}}}

class ChangeNameOp(UnaryUnaryOp):#{{{
    __slots__ = []
    def __init__(self,source, new_name):
        UnaryUnaryOp.__init__(self, source, name=new_name)#}}}

class ChangeDimOp(UnaryUnaryOp):#{{{
    __slots__ = []
    def __init__(self,source, pos, ndim):
        ndims,ntype = source.dims.updateDim(pos, ndim, source.type)
        UnaryUnaryOp.__init__(self, source, rtype=ntype,dims=ndims)#}}}

class ChangeDimPathOp(UnaryUnaryOp):#{{{
    __slots__ = []
    def __init__(self,source, ndims):
        UnaryUnaryOp.__init__(self, source, dims=ndims)#}}}

class CastOp(UnaryUnaryOp):#{{{
    __slots__ = ["cast_name"]
    def __init__(self, source, new_type):
        if(isinstance(new_type, rtypes.TypeUnknown)):
            cast_name = casts.canCast(source.type,new_type, len(source.dims))
        else:
            new_type,cast_name = casts.canCast(source.type,new_type, len(source.dims))
        assert not cast_name is False, "Cannot cast " + str(source.type) + " to " + str(new_type)
        self.cast_name = cast_name
        UnaryUnaryOp.__init__(self, source, rtype=new_type)#}}}
        
class DetectTypeOp(UnaryUnaryOp):#{{{
    __slots__ = []#}}}

class DetectAndCastOp(DetectTypeOp):#{{{
    __slots__ = ["allow_convert"]
    
    def __init__(self, source, name=None, rtype=None, dims=None, bookmarks=None, allow_convert=True):
        self.allow_convert = allow_convert
        UnaryUnaryOp.__init__(self, source,  name, rtype, dims, bookmarks)
#}}}

class DetectFixedShapesOp(UnaryUnaryOp):
    __slots__ = []


class UnpackArrayOp(UnaryUnaryOp):#{{{
    """An slice which is the result of unpacking a source slice."""
    __slots__ = ["unpack_dims"]

    def __init__(self,slice,ndim=None):
        """Creates a new slice, and sets attributes.

        :param slice: Source slice to be unpacked"""
        stype = slice.type
        if(isinstance(stype, (rtypes.TypeSet, rtypes.TypeString))):
            slice = CastOp(slice,rtypes.TypeArray)
            stype = slice.type
            

        if(ndim is None):
            if(not stype.dims):
                ndim = 1
            else:
                ndim = len(stype.dims)
        
        unpack_dims = []
        rest_dims = []
        while ndim:
            assert isinstance(stype, rtypes.TypeArray), "Cannot unpack slice deep enough " + \
                                str(slice.name) + " as there is no array type"
            
            if(not stype.dims):
                nudims = (dimensions.Dim(UNDEFINED, variable=(True,) * (len(slice.dims) + len(unpack_dims)), 
                            has_missing=stype.has_missing),)
            else: 
                nudims = stype.dims[:ndim]
                rest_dims.extend(stype.dims[ndim:])
            last_type = stype
            stype = stype.subtypes[0]
            ndim -= len(nudims)
            unpack_dims.extend(nudims)
            if(ndim and isinstance(stype, (rtypes.TypeSet, rtypes.TypeString))):
                slice = UnpackArrayOp(slice,len(unpack_dims))
                slice = CastOp(slice, rtypes.TypeArray)
                unpack_dims = []
                stype = slice.type
                res_dims = []
        
        if(rest_dims):
            ntype = rtypes.TypeArray(last_type.has_missing, dimpaths.DimPath(*rest_dims), last_type.subtypes)
        else:
            ntype = stype

        self.unpack_dims = dimpaths.DimPath(*unpack_dims)
        UnaryUnaryOp.__init__(self, slice, rtype=ntype, dims=slice.dims + self.unpack_dims)#}}}

class InsertDimOp(UnaryUnaryOp):#{{{
    __slots__ = ["matchpoint","newdim"]
    def __init__(self,slice,matchpoint,ndim):
        assert len(slice.dims) >= matchpoint, "Matchpoint for dim insertion outside dimpath"
        assert ndim.shape == 1, "Length of inserted dim should be equal to 1"
        
        ndims,ntype = slice.dims.insertDim(matchpoint,ndim, slice.type)
        self.matchpoint = matchpoint
        self.newdim = ndim
        UnaryUnaryOp.__init__(self,slice,rtype=ntype,dims=ndims)        #}}}

class EnsureCommonDimOp(UnaryUnaryOp):#{{{
    __slots__ = ["refslices","checkpos"]
    def __init__(self,slice,refslices,checkpos,bcdim):
        self.checkpos = checkpos
        self.refslices = refslices
        ndims,ntype = slice.dims.updateDim(checkpos, bcdim, slice.type)
        UnaryUnaryOp.__init__(self, slice, rtype=ntype,dims=ndims)#}}}

class BroadcastOp(UnaryUnaryOp):#{{{
    __slots__ = ["refsliceslist","plan","bcdims","partial"]
    
    def __init__(self,slice,refslices,plan,bcdims,partial=False):
        self.refsliceslist = refslices
        self.plan = plan
        self.bcdims = bcdims
        self.partial=partial

        ndims = slice.dims
        ntype = slice.type
        bcpos = 0
        for pos, planelem in enumerate(plan):
            if(planelem == BCEXIST):
                assert refslices[bcpos],"No refslices for broadcast of dim: " + str(bcdims[pos])
                bcpos += 1
                ndims, ntype = ndims.updateDim(pos, bcdims[pos], ntype)
            elif(planelem == BCCOPY or planelem == BCSOURCE):
                pass
            else:
                raise RuntimeError, "Unknown plan element in plan: " + str(plan)
            
        UnaryUnaryOp.__init__(self, slice, dims=ndims,rtype=ntype)#}}}

def broadcast(slices,mode="pos", partial=False):#{{{
    if not slices:
        return ((),())
    slicedimpaths = [s.dims for s in slices]
    if(mode == "dim"):
        nbcdims, bcdims, bcplan = dimpaths.planBroadcastMatchDim(slicedimpaths)
    elif(mode == "pos"):
        nbcdims, bcdims, bcplan = dimpaths.planBroadcastMatchPos(slicedimpaths)
    else:
        raise RuntimeError, "Unknown broadcast mode: " + str(mode)

    references = defaultdict(list)
    for bcdim,nbcdim in zip(bcdims,bcdims):
        for slice in slices:
            if bcdim in slice.dims:
                references[nbcdim].append(slice)

    nslices = []
    for plan,slice in zip(bcplan,slices):
        slice = apply_broadcast_plan(slice, plan, bcdims, references, partial)
        nslices.append(slice)

    for slice in nslices:
        #assert slice.dims == nslices[0].dims,'Dimension mismatch in broadcast: ' + str(slice.dims) + ' vs. ' + str(nslices[0].dims)
        slice.dims = dimpaths.DimPath(*nbcdims) #HACK, we have to make the dims dependent on ids
    return (nslices, bcplan)
    #}}}

def broadcast_fromplan(slice, refslices, plan, origdims, bcdims,  partial=False):#{{{
    references = defaultdict(list)
    for bcdim in bcdims:
        for rslice in refslices:
            if bcdim in rslice.dims:
                references[bcdim].append(rslice)

    sbcdims,splan = dimpaths.planBroadcastFromPlan(slice.dims, plan, origdims, bcdims)
    slice = apply_broadcast_plan(slice, splan, sbcdims, references, partial)
    return (slice, splan)
    #}}}

def apply_broadcast_plan(slice,plan, bcdims, references, partial):#{{{
    nplan = []
    active_bcdims = []
    for dimpos, bcdim, planelem in zip(range(len(bcdims)),bcdims, plan):
        if(planelem == BCNEW):
            ndim = dimensions.Dim(1)
            slice = InsertDimOp(slice,dimpos,ndim)
            active_bcdims.append(bcdim)
            nplan.append(BCEXIST)
        elif(planelem == BCENSURE):
            slice = EnsureCommonDimOp(slice,references[bcdim],dimpos,bcdim)
            nplan.append(BCSOURCE)
        elif(planelem == BCEXIST):
            active_bcdims.append(bcdim)
            nplan.append(planelem)
        else:
            nplan.append(planelem)
    if(active_bcdims):                
        slice = BroadcastOp(slice,[references[bcdim] for bcdim in active_bcdims], nplan, bcdims, partial)
    return slice
#}}}

def broadcastParentsFromPlan(slice, seldimpath, plan, origdims, bcdims, refslices, partial):#{{{
    used_dims = [False,] * len(slice.dims)
    if not plan:
        return slice
    while True:
        #determine filter dim
        lastpos = slice.dims.matchDimPath(seldimpath)
        for filterpos in lastpos[::-1]:
             if(not used_dims[filterpos]):
                break
        else:
            break

        #pack up to and including filter dim
        packdepth = len(slice.dims) - filterpos
        if(packdepth):
            slice = PackArrayOp(slice,packdepth)

        #broadcast to constraint
        slice,splan = broadcast_fromplan(slice, refslices, plan, origdims, bcdims, partial)

        #adapt used_dims
        used_dims = dimpaths.applyPlan(used_dims,splan,newvalue=True,copyvalue=True,ensurevalue=True,existvalue=True)
        used_dims[filterpos] = True

        #unpack dims
        if(packdepth):
            slice = UnpackArrayOp(slice,packdepth)
            
    return slice#}}}

class PermuteDimsOp(UnaryUnaryOp):#{{{
    __slots__ = ["permute_idxs"]
    
    def __init__(self,slice,permute_idxs):
        assert len(permute_idxs) == len(slice.dims), "Number of permute indexes not equal to number of dimensions"
        
        ndims,ntype = slice.dims.permuteDims(permute_idxs,slice.type)
        self.permute_idxs = permute_idxs
        UnaryUnaryOp.__init__(self,slice, dims=ndims,rtype=ntype)#}}}

class SplitDimOp(UnaryUnaryOp):#{{{
    __slots__ = ["pos","lshape","rshape"]

    def __init__(self,slice, pos,lshape,rshape,ldim,rdim):
        npath,ntype = slice.dims.updateDim(pos, (ldim,rdim),slice.type)
        self.pos = pos
        self.lshape = lshape
        self.rshape = rshape

        UnaryUnaryOp.__init__(self,slice,dims=npath,rtype=ntype)#}}}

class ShapeOp(UnaryUnaryOp):#{{{
    __slots__ = ["pos"]
    def __init__(self,slice,pos):
        d = slice.dims[pos]
        ntype = rtypes.unknown
        self.pos = pos
        UnaryUnaryOp.__init__(self,slice, name=d.name,rtype=ntype,dims=dimpaths.DimPath())#}}}

class FilterOp(UnaryUnaryOp):#{{{
    __slots__ = ["constraint","has_missing"]

    def __init__(self,slice,constraint, ndim,refconstraint=None):
        if refconstraint is None:
            refconstraint = constraint

        stype = slice.type
        assert isinstance(stype, rtypes.TypeArray), "Filter on non-array type not possible"
        
        if(isinstance(constraint.type,rtypes.TypeArray)):
            has_missing = constraint.type.subtypes[0].has_missing
        else:
            has_missing = constraint.type.has_missing
        self.has_missing = has_missing            

        sdims = stype.dims
        if(ndim is None):
            sdims, subtype = sdims.removeDim(0, refconstraint, stype.subtypes[0])
        else:
            sdims, subtype = sdims.updateDim(0, ndim, stype.subtypes[0])
        
        
        if(sdims):
            ntype = dimpaths.dimsToArrays(sdims, subtype)
        else:
            ntype = subtype
        
        if(has_missing):
            ntype = ntype.setHasMissing(True)
            if not subtype.hasMissingValInfo():
                slice = DetectFixedShapesOp(slice)
      
        self.constraint = constraint
        UnaryUnaryOp.__init__(self, slice, rtype=ntype)#}}}

def filter(slice,constraint,seldimpath, ndim, mode="dim"):#{{{
    used_dims = [False,] * len(slice.dims)
    while True:
        #determine filter dim
        lastpos = slice.dims.matchDimPath(seldimpath)

        for filterpos in lastpos[::-1]:
             if(not used_dims[filterpos]):
                break
        else:
            break

        #pack up to and including filter dim
        packdepth = len(slice.dims) - filterpos
        if(packdepth):
            tslice = PackArrayOp(slice,packdepth)
        else:
            tslice = slice


        #prepare adaptation of ndim.dependent
        if(not ndim is None):
            dep = list(ndim.dependent)
            while(len(dep) < len(constraint.dims)):
                dep.insert(0,False)

        #broadcast to constraint
        (tslice,tconstraint),(splan,cplan) = broadcast([tslice,constraint],mode=mode, partial=True)

        #adapt ndim to braodcast, apply filter
        #FIXME: change dependent dims of slice.type

        if(not ndim is None):
            if(isinstance(tconstraint.type,rtypes.TypeSlice) and seldimpath[-1].dependent): #if variable, extend variability
                ndep = dimpaths.applyPlan(dep[::-1],cplan,newvalue=True, copyvalue=True, ensurevalue=True)
            else:
                ndep = dimpaths.applyPlan(dep[::-1],cplan,newvalue=False)
            xndim = ndim.changeDependent(tuple(ndep[::-1]), tslice.dims)
        else:
            #optimization: do not broadcast array of integer dim, to later remove it
            if cplan and cplan[-1] == BCSOURCE:
                ndim = constraint.dims[-1]
                constraint = PackArrayOp(constraint)
                continue
            xndim = ndim
        slice = FilterOp(tslice,tconstraint,xndim,constraint)

        #adapt used_dims
        used_dims = dimpaths.applyPlan(used_dims,splan,newvalue=True,copyvalue=True,ensurevalue=True, existvalue=True)
        
        #handle dim removal, filter_dim is used
        if(ndim is None): #dimension removed/collapsed
            packdepth -= 1
            del used_dims[len(cplan)]
        else:
            used_dims[len(cplan)] = True

        #unpack dims
        if(packdepth):
            slice = UnpackArrayOp(slice,packdepth)

        #break out of loop if there was only one dim match (necessary to prevent 
        #bug in which a dimpath starting with root gets matched multiple times)
        if len(lastpos) <= 1:
            break

    return slice#}}}

class FlatAllOp(UnaryUnaryOp):#{{{
    __slots__ = []

    def __init__(self,slice, ndim):
        stype = slice.type
        sdims = slice.dims
        for i in xrange(len(sdims)):
            stype = stype._removeDepDim((i+1), None)
        stype = stype._insertDepDim(1,ndim)
        dims = dimpaths.DimPath(ndim)

        UnaryUnaryOp.__init__(self, slice, rtype=stype,dims=dims)#}}}

class FlatDimOp(UnaryUnaryOp):#{{{
    __slots__ = ["flatpos"]

    def __init__(self,slice,flatpos, ndim):
        stype = slice.type
        sdims = slice.dims
        sdims,stype = sdims.removeDim(flatpos,None,stype)
        sdims,stype = sdims.updateDim(flatpos-1,ndim,stype)

        self.flatpos = flatpos
        UnaryUnaryOp.__init__(self, slice, rtype=stype,dims=sdims)#}}}

class UnpackTupleOp(UnaryUnaryOp):#{{{
    """A slice which is the result of unpacking a tuple slice."""
    __slots__ = ["tuple_idx"]
    
    def __init__(self, slice, idx):
        """Creates a new slice, using source `slice`, by 
        extracting the `idx` subtype.
        
        :param slice: new slice
        :param idx:   index of tuple attribute to be unpacked"""
        
        stype = slice.type
        assert isinstance(stype, rtypes.TypeTuple), "Cannot unpack slice " + \
                                str(slice.name) + " as it is not a tuple"
        
        if not 0 <= idx < len(slice.type.subtypes):
            raise RuntimeError, "Tuple index invalid, outside range of available attributes"

        ntype = slice.type.subtypes[idx]
        
        if(slice.type.fieldnames):
            name = slice.type.fieldnames[idx]
        else:
            name = "f" + str(idx)

        UnaryUnaryOp.__init__(self, slice, name=name, rtype=ntype)
        self.tuple_idx = idx#}}}

class PackArrayOp(UnaryUnaryOp):#{{{
    __slots__ = ["pack_dims"]

    def __init__(self, pslice, ndim=1):
        assert len(pslice.dims) >= ndim, "Op does not have enough dimensions to pack as " + str(ndim) + "-dimensional array"
        
        dims = pslice.dims[-ndim:]
        self.pack_dims = dims
        ntype = dimpaths.dimsToArrays(dims,pslice.type)
        UnaryUnaryOp.__init__(self, pslice, rtype=ntype, dims=pslice.dims[:-ndim])#}}}

class PackListOp(PackArrayOp):#{{{
    __slots__ = []

    def __init__(self, pslice):
        PackArrayOp.__init__(self, pslice, 1)
        #}}}

class ToPythonOp(UnaryUnaryOp):#{{{
    __slots__ = []

    def __init__(self, pslice):
        ntype = pslice.type.copy()
        UnaryUnaryOp.__init__(self, pslice, rtype=ntype)#}}}
        
class NoneToMissingOp(UnaryUnaryOp):#{{{
    __slots__ = []
    def __init__(self, slice):
        assert slice.type.has_missing, "Cannot apply none to missing on type without has_missing"
        UnaryUnaryOp.__init__(self, slice)#}}}
    
class FreezeOp(UnaryUnaryOp):#{{{
    __slots__ = []
#}}}

def ensure_frozen(slice):#{{{
    if(freeze_protocol.needFreeze(slice.type)):
        return FreezeOp(slice)
    else:
        return slice#}}}

class UnaryFuncOp(UnaryUnaryOp):#{{{
    __slots__ = ["funcname","sig", "kwargs"]
    def __init__(self, slice, funcname, sig, outparam, dims=None, **kwargs):
        self.funcname = funcname
        self.sig = sig
        self.kwargs = kwargs

        if(dims is None):
            dims = slice.dims
        UnaryUnaryOp.__init__(self, slice, name=outparam.name, rtype=outparam.type, dims=dims)#}}}

class UnaryFuncElemOp(UnaryFuncOp):#{{{
    __slots__ = []
    
    def __init__(self, funcname, sig, outparam, slice, **kwargs):
        UnaryFuncOp.__init__(self, slice, funcname, sig, outparam, **kwargs)
        #}}}

class UnaryFuncSeqOp(UnaryFuncOp):#{{{
    __slots__ = ["packdepth"]

    
    def __init__(self, funcname, sig, outparam, packdepth, slice, **kwargs):
        self.packdepth = packdepth
        UnaryFuncOp.__init__(self, slice, funcname, sig, outparam, **kwargs)
        #}}}

class UnaryFuncAggregateOp(UnaryFuncOp):#{{{
    __slots__ = ["packdepth"]
    
    def __init__(self, funcname, sig, outparam, packdepth, slice, **kwargs):
        self.packdepth = packdepth
        sdims, ntype = slice.dims.removeDim(len(slice.dims) - packdepth, (funcname,outparam), outparam.type)

        UnaryFuncOp.__init__(self, slice, funcname, sig, outparam, dims=sdims, **kwargs)
        self.type = ntype
        #}}}

class MultiUnaryOp(UnaryOp):#{{{
    __slots__ = ["sources"]
    def __init__(self, source_slices, name, rtype=rtypes.unknown, 
                            dims=dimpaths.DimPath(), bookmarks=set()):
        self.sources = tuple(source_slices)
        UnaryOp.__init__(self, name, rtype, dims, bookmarks)#}}}

class BinFuncOp(MultiUnaryOp):#{{{
    __slots__ = ["funcname","sig"]
    def __init__(self, lslice, rslice, funcname, sig, outparam, dims=None):
        self.funcname = funcname
        self.sig = sig

        if(dims is None):
            assert lslice.dims == rslice.dims, "Op dims not equal, and no dim given"
            dims = lslice.dims
        nbookmarks = lslice.bookmarks | rslice.bookmarks

        MultiUnaryOp.__init__(self, (lslice, rslice), name=outparam.name, rtype=outparam.type, dims=dims, bookmarks=nbookmarks)#}}}

class BinFuncElemOp(BinFuncOp):#{{{
    def __init__(self, funcname, sig, outparam, left, right):
        dims = left.dims
        assert all([d1 == d2 for d1, d2 in zip(dims, right.dims)]), \
                    "Dimensions of slices do not match"
        
        BinFuncOp.__init__(self, left, right, funcname, sig, outparam, dims)#}}}

class PackTupleOp(MultiUnaryOp):#{{{
    __slots__ = []
    ocls = rtypes.TypeTuple

    def __init__(self, slices, field="data"):
        cdim = set([slice.dims for slice in slices])
        assert len(cdim) == 1, "Packing tuple on slices with different dims"
        
        fieldnames = [slice.name for slice in slices]
        subtypes = [slice.type for slice in slices]
        ntype = self.ocls(False, tuple(subtypes), tuple(fieldnames))
        nbookmarks = reduce(set.union,[slice.bookmarks for slice in slices])
        MultiUnaryOp.__init__(self, slices, name=field, rtype=ntype, dims=iter(cdim).next(),bookmarks=nbookmarks)#}}}

class PackDictOp(PackTupleOp):#{{{
    __slots__ = ["with_missing"]
    ocls = rtypes.TypeRecordDict

    def __init__(self, slices, field="data", with_missing=False):
        self.with_missing = with_missing
        PackTupleOp.__init__(self, slices, field)#}}}


class PackIndexDictOp(MultiUnaryOp):#{{{
    __slots__ = []
    ocls = rtypes.TypeIndexDict

    def __init__(self, slices, field="data"):
        assert len(slices) == 2, "Index dict op should have two slices"
        assert slices[0].dims == slices[1].dims, "Key and value of dict should have same dimension"

        slices = [PackArrayOp(slice) for slice in slices]
        ndims = dimpaths.DimPath(dimensions.Dim(UNDEFINED, (), name="dict"))
        subtypes = [dimpaths.dimsToArrays(ndims, slice.type.subtypes[0]) for slice in slices]
        
        fieldnames = [slice.name for slice in slices]
        ntype = self.ocls(False, tuple(subtypes), tuple(fieldnames))
        nbookmarks = reduce(set.union,[slice.bookmarks for slice in slices])
        MultiUnaryOp.__init__(self, slices, name=field, rtype=ntype, dims=slices[0].dims, bookmarks=nbookmarks)#}}}
        #}}}


class TakeOp(MultiUnaryOp):
    __slots__ = ['allow_missing','keep_missing']
    def __init__(self, source_slice, take_slice, allow_missing=False, keep_missing=False):
        assert isinstance(source_slice.type, rtypes.TypeIndexDict), "Take left op should have type index dict"
        source_slice, take_slice = broadcast((source_slice, take_slice), mode="dim")[0]
        ntype = source_slice.type.subtypes[1].subtypes[0]
        if allow_missing:
            ntype = ntype.setHasMissing(True)
        elif keep_missing:
            ntype = casts.castImplicitCommonType(ntype,take_slice.type)

        if keep_missing:
            name = take_slice.name
        else:
            name = source_slice.type.fieldnames[1]
        nbookmarks = source_slice.bookmarks | take_slice.bookmarks
        self.allow_missing = allow_missing
        self.keep_missing = keep_missing
        MultiUnaryOp.__init__(self, [source_slice, take_slice], name=name, rtype=ntype, dims=take_slice.dims, bookmarks = nbookmarks)#}}}

class GroupIndexOp(MultiUnaryOp):#{{{
    def __init__(self, slices):
        assert len(set([slice.dims for slice in slices])) == 1, "Group index slices should have same dimension"
        assert all([isinstance(slice.type,rtypes.TypeArray) for slice in slices]), "Group index slices should be arrays"
        ndims = slices[0].dims
         
        newdims = []
        dep = (True,) * len(ndims)
        for pos, slice in enumerate(slices):
            xdep = (False,) * pos + dep
            newdims.append(dimensions.Dim(UNDEFINED,dependent=xdep,name="g" + slice.name))

        sdim = dimensions.Dim(UNDEFINED,(True,) * (len(ndims) + len(newdims)),name="g" + slices[0].type.dims[0].name)
        stype = rtypes.TypeInt64()
        rtype = dimpaths.dimsToArrays(dimpaths.DimPath(sdim), stype)
        rtype = dimpaths.dimsToArrays(dimpaths.DimPath(*newdims), rtype)
        MultiUnaryOp.__init__(self,slices, name="groupindex", rtype=rtype, dims=ndims)#}}}

class HArrayOp(MultiUnaryOp):#{{{
    __slots__ = []

    def __init__(self, slices, name=None):
        cdim = set([slice.dims for slice in slices])
        assert len(cdim) == 1, "Packing tuple on slices with different dims"
        
        subtypes = [slice.type for slice in slices]
        assert len(set(subtypes)) == 1, "HArray can only be applied if types are equal"

        ndim = dimensions.Dim(len(slices), name=name)
        ntype = rtypes.TypeArray(False, dimpaths.DimPath(ndim), (subtypes[0],))

        nbookmarks = reduce(set.union,[slice.bookmarks for slice in slices])

        slicename = "_".join([slice.name for slice in slices])
        MultiUnaryOp.__init__(self, slices, name=slicename, rtype=ntype, dims=iter(cdim).next(),bookmarks=nbookmarks)#}}}

class FixateOp(MultiUnaryOp):
    __slots__ = []
    def __init__(self, slices):
        MultiUnaryOp.__init__(self, slices, "result")

class GatherOp(FixateOp):
    __slots__ = []

class MultiOp(Op):
    __slots__ = ["results"]
    
    def __init__(self, results):
        assert isinstance(results, tuple), "Results should be a tuple"
        self.results = results

class SelectOp(UnaryUnaryOp):#{{{
    __slots__ = ["index"]
    def __init__(self, slice, index, name, rtype, ndims, bookmarks):
        assert isinstance(slice, MultiOp), "Source lice of SelectOp should be a multiop"
        self.index = index
        UnaryUnaryOp.__init__(self, slice, name=name, rtype=rtype, dims=ndims, bookmarks=bookmarks)#}}}

class UnaryMultiOp(MultiOp):
    __slots__ = ["source"]

class MultiMultiOp(MultiOp):#{{{
    __slots__ = ["sources"]

    def __init__(self, sources, results):
        assert isinstance(sources, tuple), "Sources should be a tuple"
        self.sources = sources
        MultiOp.__init__(self, results)#}}}


class EquiJoinIndexOp(MultiMultiOp):#{{{
    __slots__ = ["jointype"]

    def __init__(self, leftslice, rightslice, jointype="inner", mode="dim"):
        assert leftslice.dims == rightslice.dims, "Parent dims of join fields should be equal"
        if leftslice.type.dims[0].name == rightslice.type.dims[0].name:
            name= leftslice.type.dims[0].name
        else:
            name= leftslice.type.dims[0].name + "_" + rightslice.type.dims[0].name
        ndim = dimensions.Dim(UNDEFINED, (True,) * len(leftslice.dims), name=name)

        ltype = rtypes.TypePlatformInt(has_missing=(jointype == "right" or jointype=="full"))
        ltype = rtypes.TypeArray(subtypes=(ltype,), dims=dimpaths.DimPath(ndim))
        
        rtype = rtypes.TypePlatformInt(has_missing=(jointype == "left" or jointype=="full"))
        rtype = rtypes.TypeArray(subtypes=(rtype,), dims=dimpaths.DimPath(ndim))

        r1 = SelectOp(self,0,leftslice.name, ltype,leftslice.dims, leftslice.bookmarks)
        r2 = SelectOp(self,1,rightslice.name, rtype,leftslice.dims, rightslice.bookmarks)
        self.jointype = jointype
        MultiMultiOp.__init__(self, (leftslice,rightslice),(r1,r2))#}}}




